"""Rattr function analyser."""

import ast
from itertools import accumulate
from typing import List, Optional, Tuple

from rattr import error
from rattr.analyser.base import NodeVisitor
from rattr.analyser.context import (
    Call,
    Class,
    Context,
    Func,
    Name,
    Symbol,
    new_context,
)
from rattr.analyser.types import (
    AnyAssign,
    AnyFunctionDef,
    CompoundStrictlyNameable,
    FunctionIR,
    Nameable,
    StrictlyNameable,
    ast_NamedExpr,
)
from rattr.analyser.util import (
    LOCAL_VALUE_PREFIX,
    PYTHON_ATTR_BUILTINS,
    assignment_is_one_to_one,
    class_in_rhs,
    get_assignment_targets,
    get_basename,
    get_basename_fullname_pair,
    get_fullname,
    get_function_body,
    get_function_call_args,
    get_function_def_args,
    is_call_to,
    lambda_in_rhs,
    remove_call_brackets,
)
from rattr.plugins import plugins


class FunctionAnalyser(NodeVisitor):
    """Walk a function's AST to determine the accessed attributes."""

    def __init__(self, _ast: ast.AST, context: Context) -> None:
        """Set configuration and initialise IR."""
        if not isinstance(_ast, AnyFunctionDef.__args__):
            raise TypeError("FunctionAnalyser expects `_ast` to be a function")

        self._ast: AnyFunctionDef = _ast

        self.func_ir: FunctionIR = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }

        # NOTE Managed by `new_context` contextmanager -- do not manually set
        self.context: Context = context

    def analyse(self) -> FunctionIR:
        """Entry point, return the results of analysis."""
        with new_context(self):
            self.context.push_arguments_to_context(self._ast.args)

            for stmt in get_function_body(self._ast):
                self.visit(stmt)

        return self.func_ir

    def get_and_verify_name(
        self, node: Nameable, ctx: ast.expr_context
    ) -> Tuple[str, str]:
        """Return the name, also verify validity."""
        base, full = get_basename_fullname_pair(node, safe=True)

        is_undeclared = base not in self.context
        is_assignment = isinstance(ctx, ast.Store)
        is_local = base.startswith(LOCAL_VALUE_PREFIX)

        if is_undeclared and not is_assignment and not is_local:
            error.warning(f"'{base}' potentially undefined", node)

        return base, full

    def update_results(self, symbol: Symbol, ctx: ast.expr_context) -> None:
        """Add the given name to the result.

        To be called on access -- i.e. ast.Name, etc.

        """
        if not isinstance(symbol, Symbol):
            raise TypeError()

        if isinstance(ctx, ast.Store):
            self.func_ir["sets"].add(symbol)

        if isinstance(ctx, ast.Load):
            self.func_ir["gets"].add(symbol)

        if isinstance(ctx, ast.Del):
            self.func_ir["dels"].add(symbol)

    # ----------------------------------------------------------------------- #
    # Result alterors: variables, call expressions, and subscripting
    # ----------------------------------------------------------------------- #

    def visit_Name(self, node: ast.Name) -> None:
        """Visit ast.Name(id: str, ctx: ast.expr_context)."""
        basename, fullname = self.get_and_verify_name(node, node.ctx)

        self.update_results(Name(fullname, basename), node.ctx)

    def visit_compound_nameable(self, node: CompoundStrictlyNameable) -> None:
        """Helper method for special nameable nodes.

        Visit ast.Starred(value, ctx).
        Visit ast.Attribute(value, attr: str, ctx: ast.expr_context).
        Visit ast.Subscript(value, slice, ctx).

        """
        basename, fullname = self.get_and_verify_name(node, node.ctx)

        # NOTE
        #   Visit the operands in a BinOp, etc.
        #   I.e.:
        #       In `(a + b).thing`, the the names `a` and `b` should also be
        #       visited
        #       In `a.thing`, the expr should be visted once as a whole
        #       (neither `a` nor `thing` should be visited directly)
        if not isinstance(node.value, StrictlyNameable.__args__):
            self.generic_visit(node.value)

        self.update_results(Name(fullname, basename), node.ctx)

    def visit_Starred(self, node: ast.Starred) -> None:
        self.visit_compound_nameable(node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        self.visit_compound_nameable(node)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        self.visit_compound_nameable(node)

    def visit_Call(self, node: ast.Call) -> None:
        """Visit ast.Call(func, args, keywords)."""
        _, fullname = self.get_and_verify_name(node, ast.Load())

        # Handle special builtins such as getattr, setattr, etc
        if self.handle_special_function(node):
            return

        # NOTE
        #   Add the call to the IR and manually visit the arguments,
        #   but not `node.func` as it will register an incorrect "gets"

        target = self.context.get_call_target(fullname, node)

        if not isinstance(target, Class):
            args = get_function_call_args(node)
        else:
            error.warning(f"'{target.name}' initialised but not stored", node)
            args = get_function_call_args(node, LOCAL_VALUE_PREFIX + target.name)

        # NOTE
        #   If this is a call to a method on an attribute, then it necessarily
        #   "gets" the attribute
        parts = remove_call_brackets(fullname).split(".")[:-1]
        for attr in list(accumulate(parts, lambda a, b: f"{a}.{b}"))[1:]:
            self.func_ir["gets"].add(Name(attr, parts[0]))

        self.func_ir["calls"].add(Call(fullname, *args, target))

        for arg in (*node.args, *node.keywords):
            self.visit(arg)

    # ----------------------------------------------------------------------- #
    # Result alterors: special functions
    # ----------------------------------------------------------------------- #

    def handle_special_function(self, node: ast.Call) -> bool:
        """Return `True` if handled."""
        if isinstance(node.func, ast.Name):
            name = remove_call_brackets(node.func.id)
        else:
            name = remove_call_brackets(get_fullname(node, safe=True))

        analyser = plugins.custom_function_handler.get(name, self.context.get_root())

        if analyser is None:
            return False

        ir = analyser.on_call(name, node, self.context)

        self.func_ir["sets"] = set.union(self.func_ir["sets"], ir["sets"])
        self.func_ir["gets"] = set.union(self.func_ir["gets"], ir["gets"])
        self.func_ir["dels"] = set.union(self.func_ir["dels"], ir["dels"])
        self.func_ir["calls"] = set.union(self.func_ir["calls"], ir["calls"])

        return True

    # ----------------------------------------------------------------------- #
    # Context alterors: assignments and deleteion
    # ----------------------------------------------------------------------- #

    def visit_LambdaAssign(self, node: AnyAssign, targets: List[ast.expr]) -> None:
        """Helper method for handling non-anonymous lambdas."""
        target = targets[0]

        if not assignment_is_one_to_one(node):
            error.fatal("lambda assignment must be one-to-one", node)

        error.error("unable to unbind lambdas defined in functions", node)

        name = get_fullname(target)
        func = Func(name, *get_function_def_args(node.value))

        self.context.add(func)

    def visit_ClassAssign(self, node: AnyAssign, targets: List[ast.expr]) -> None:
        """Helper method for assignments where RHS is a new class instance."""
        target = targets[0]

        if not assignment_is_one_to_one(node):
            error.fatal("class assignment must be one-to-one", node)

        # Create call to class initialiser
        lhs_basename, lhs_name = get_basename_fullname_pair(target)
        class_name = get_fullname(node.value)

        init_args = get_function_call_args(node.value, lhs_name)
        init_body = self.context.get_call_target(class_name, node)

        self.func_ir["calls"].add(Call(class_name, *init_args, init_body))

        # Create set to LHS
        self.func_ir["sets"].add(Name(lhs_name, lhs_basename))

        # Register assignments
        for target in targets:
            self.context.add_identifiers_to_context(target)

        # Visit call arguments
        for arg in (*node.value.args, *node.value.keywords):
            self.visit(arg)

    def visit_AnyAssign(self, node: AnyAssign) -> None:
        """Helper method for assignments.

        Visit ast.Assign(targets, value, type_comment)\\
        Visit ast.AnnAssign(target, annotation, value, simple)\\
        Visit ast.AugAssign(target, op, value)

        """
        targets = get_assignment_targets(node)

        # NOTE Handle special case, non-anonymous lambda
        if lambda_in_rhs(node):
            self.visit_LambdaAssign(node, targets)
            return

        # NOTE Handle special case, rhs is class
        if class_in_rhs(node, self.context):
            self.visit_ClassAssign(node, targets)
            return

        for target in targets:
            self.context.add_identifiers_to_context(target)

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast_NamedExpr) -> None:
        self.func_ir["sets"].add(Name(*get_basename_fullname_pair(node.target)))

        if lambda_in_rhs(node):
            self.generic_visit(node.value)

        self.visit_AnyAssign(node)

    def visit_Delete(self, node: ast.Delete) -> None:
        """Visit ast.Delete(targets)."""
        for target in node.targets:
            self.context.del_identifiers_from_context(target)

        return self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        """Visit ast.For(target, iter, body, orelse, type_comment)."""
        self.context.add_identifiers_to_context(node.target)

        return self.generic_visit(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        """Visit ast.AsyncFor(target, iter, body, orelse)."""
        self.context.add_identifiers_to_context(node.target)

        return self.generic_visit(node)

    def visit_With(self, node: ast.With) -> None:
        """Visit ast.With(items, body, type_comment)."""
        for item in node.items:
            if not item.optional_vars:
                continue
            self.context.add_identifiers_to_context(item.optional_vars)

        return self.generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        """Visit ast.class AsyncWith(items, body)."""
        for item in node.items:
            if not item.optional_vars:
                continue
            self.context.add_identifiers_to_context(item.optional_vars)

        return self.generic_visit(node)

    # ----------------------------------------------------------------------- #
    # Context creators: function, class definitions, etc
    # ----------------------------------------------------------------------- #

    def visit_AnyFunctionDef(self, node: AnyFunctionDef) -> None:
        """Helper method for visiting nested function definitions.

        Visit ast.FunctionDef(name, args, body, decorator_list, returns)\\
        Visit ast.AsyncFunctionDef(name, args, body, decorator_list, returns)\\
        Visit ast.Lambda(args, body)

        NOTE:
            visit_Lambda(lambda) and thus visit_AnyFunctionDef(lambda), are
            only reached when the lambda is an anonymous function.

        """
        args = get_function_def_args(node)

        if not isinstance(node, ast.Lambda):
            error.error("unable to unbind nested functions", node)
        else:
            error.error("unable to unbind anonymous lambdas", node)

        if isinstance(node, ast.FunctionDef):
            self.context.add(Func(node.name, *args))
        if isinstance(node, ast.AsyncFunctionDef):
            self.context.add(Func(node.name, *args, is_async=True))

        # TODO Allow nested -- FunctionAnalyser on name "outer.inner"
        with new_context(self):
            self.context.push_arguments_to_context(node.args)

            for stmt in get_function_body(node):
                self.visit(stmt)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        return error.error("nested classes unsupported", node)

    # ----------------------------------------------------------------------- #
    # Context creators: comprehensions and generators
    # ----------------------------------------------------------------------- #

    def visit_comprehension(self, node: ast.comprehension) -> None:
        """Visit ast.comprehension(target, iter, ifs, is_async)."""
        # Add target to context
        self.context.add_identifiers_to_context(node.target)

        # If iter isn't defined by another target, visit it
        basename = get_basename(node.iter, safe=True)
        if not self.context.declares(basename):
            self.generic_visit(node.iter)

        for if_stmt in node.ifs:
            self.generic_visit(if_stmt)

    def visit_GenericComp(
        self, generators: List[ast.comprehension], exprs: List[ast.expr]
    ) -> None:
        """Helper method visit an arbitrary comprehension.

        Visit ast.ListComp(elt: ast.expr, generators: ast.comprehension).
        Visit ast.SetComp(elt, generators).
        Visit ast.GeneratorExp(elt, generators).
        Visit ast.DictComp(key, value, generators).

        NOTE:
            Ordinarily, the attributes are visited in order:
                elt / key, value
                generators
            However, as the generators create a local scope, we wish to visit
            them first.

        """
        with new_context(self):
            for comp in generators:
                self.visit_comprehension(comp)

            for expr in exprs:
                self.generic_visit(expr)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        self.visit_GenericComp(node.generators, [node.elt])

    def visit_SetComp(self, node: ast.SetComp) -> None:
        self.visit_GenericComp(node.generators, [node.elt])

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        self.visit_GenericComp(node.generators, [node.elt])

    def visit_DictComp(self, node: ast.DictComp) -> None:
        self.visit_GenericComp(node.generators, [node.key, node.value])

    # ----------------------------------------------------------------------- #
    # Special cases
    # ----------------------------------------------------------------------- #

    def visit_ReturnValue(self, node: Optional[ast.expr]) -> bool:
        """Helper method to handle a return value / tuple elt.

        Returns True if the return value has been fully handled.

        """
        if node is None:
            return True

        if isinstance(node, (ast.Tuple, ast.List, ast.Set)):
            for elt in node.elts:
                handled = self.visit_ReturnValue(elt)

                if not handled:
                    self.visit(elt)

            return True

        if isinstance(node, ast.Dict):
            for elt in (*node.keys, *node.values):
                handled = self.visit_ReturnValue(elt)

                if not handled:
                    self.visit(elt)

            return True

        if isinstance(node, ast.Call):
            # NOTE
            #   get_basename_fullname_pair on getattr, etc. produces a name
            #   incompatible with Context::get_call_target
            if any(is_call_to(f, node) for f in PYTHON_ATTR_BUILTINS):
                return False

            target = self.context.get_call_target(
                get_fullname(node, True), node, warn=False
            )

            if not isinstance(target, Class):
                return False

            # Create call to class initialiser
            class_name = get_fullname(node)

            init_args = get_function_call_args(node, "@ReturnValue")
            init_body = self.context.get_call_target(class_name, node)

            self.func_ir["calls"].add(Call(class_name, *init_args, init_body))

            # Visit call arguments
            for arg in (*node.args, *node.keywords):
                self.visit(arg)

            return True

        return False

    def visit_Return(self, node: ast.Return) -> None:
        """Visit ast.Return(value)."""
        handled = self.visit_ReturnValue(node.value)

        if not handled:
            self.visit(node.value)

    # ----------------------------------------------------------------------- #
    # THE FORBIDEN ZONE: A zone... that is, yes... FORBIDDEN to you.
    # ----------------------------------------------------------------------- #

    def visit_Global(self, node: ast.Global) -> None:
        return error.fatal("do not use global keyword", node)

    def visit_Nonlocal(self, node: ast.Nonlocal) -> None:
        return error.fatal("do not use nonlocal keyword", node)

    def visit_Import(self, node: ast.Import) -> None:
        return error.fatal("imports must be at the top level", node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        return error.fatal("imports must be at the top level", node)



"""Rattr Base class tests."""

from unittest import mock

import pytest

from rattr.analyser.base import (  # CustomFunctionAnalyser,
    Assertor,
    CustomFunctionHandler,
)
from rattr.analyser.context.symbol import Call, Import, Name


class TestAssertor:
    def test_assertor_is_strict(self, capfd):
        assertor = Assertor(is_strict=True)

        with mock.patch("sys.exit") as _exit:
            assertor.failed("the reason")

        output, _ = capfd.readouterr()

        assert "the reason" in output
        assert _exit.call_count == 1

    def test_assertor_not_is_strict(self, capfd):
        assertor = Assertor(is_strict=False)

        with mock.patch("sys.exit") as _exit:
            assertor.failed("the reason")

        output, _ = capfd.readouterr()

        assert "the reason" in output
        assert _exit.call_count == 0


class TestCustomFunctionHandler:
    def test_register_builtin(self, PrintBuiltinAnalyser):
        handler = CustomFunctionHandler()

        assert len(handler._builtins) == 0

        handler = CustomFunctionHandler([PrintBuiltinAnalyser()], [])

        assert len(handler._builtins) == 1

    def test_register_user_def(self, ExampleFuncAnalyser):
        handler = CustomFunctionHandler()

        assert len(handler._user_def) == 0

        handler = CustomFunctionHandler([], [ExampleFuncAnalyser()])

        assert len(handler._user_def) == 1

    def test_get(self, handler: CustomFunctionHandler):
        ctx = mock.Mock()
        ctx.is_import.return_value = True
        ctx.symbol_table.symbols.return_value = []

        assert handler.get("print", ctx).name == "print"
        assert handler.get("example", ctx).name == "example"

        assert handler.get("some_non_sense", ctx) is None

        ctx.symbol_table.symbols.return_value = [
            Import("exy", "module.example"),
            Import("mod", "module"),
        ]

        assert handler.get("exy", ctx).name == "example"
        assert handler.get("mod.example", ctx).name == "example"

        assert handler.get("some_non_sense", ctx) is None

    def test_has_analyser(self, handler: CustomFunctionHandler):
        ctx = mock.Mock()
        ctx.is_import.return_value = True
        ctx.symbol_table.symbols.return_value = []

        assert handler.has_analyser("print", ctx)
        assert handler.has_analyser("example", ctx)

        assert not handler.has_analyser("wibble", ctx)
        assert not handler.has_analyser("wibble_wibble", ctx)

    def test_handle_def(self, handler: CustomFunctionHandler):
        ctx = mock.Mock()
        ctx.is_import.return_value = True
        ctx.symbol_table.symbols.return_value = []

        call = mock.Mock()

        with pytest.raises(ValueError):
            handler.handle_def("wibble", call, ctx)

        assert handler.handle_def("print", call, ctx) == {
            "sets": {
                Name("set_in_print_def"),
            },
            "gets": {
                Name("get_in_print_def"),
            },
            "dels": {
                Name("del_in_print_def"),
            },
            "calls": {
                Call("call_in_print_def", [], {}, None),
            },
        }

        assert handler.handle_def("example", call, ctx) == {
            "sets": {
                Name("set_in_example_def"),
            },
            "gets": {
                Name("get_in_example_def"),
            },
            "dels": {
                Name("del_in_example_def"),
            },
            "calls": {
                Call("call_in_example_def", [], {}, None),
            },
        }

    def test_handle_call(self, handler: CustomFunctionHandler):
        ctx = mock.Mock()
        ctx.is_import.return_value = True
        ctx.symbol_table.symbols.return_value = []

        call = mock.Mock()

        with pytest.raises(ValueError):
            handler.handle_call("wibble", call, ctx)

        assert handler.handle_call("print", call, ctx) == {
            "sets": {
                Name("set_in_print"),
            },
            "gets": {
                Name("get_in_print"),
            },
            "dels": {
                Name("del_in_print"),
            },
            "calls": {
                Call("call_in_print", [], {}, None),
            },
        }

        assert handler.handle_call("example", call, ctx) == {
            "sets": {
                Name("set_in_example"),
            },
            "gets": {
                Name("get_in_example"),
            },
            "dels": {
                Name("del_in_example"),
            },
            "calls": {
                Call("call_in_example", [], {}, None),
            },
        }



from unittest import mock

from rattr.analyser.context import RootContext
from rattr.plugins.assertors.import_clobbering import ImportClobberingAssertor


class TestImportClobberingAssertor:
    def test_allow_non_clobbering(self, parse):
        _ast = parse(
            """
            from math import pi as pie

            pi = 3.2 - 0.06 + 0.001 # ...
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        assert not _exit.called

    def test_assign_to_imported_name(self, parse, capfd):
        # In root context
        _ast = parse(
            """
            import math

            math = "this is very bad"
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # In child context
        _ast = parse(
            """
            from os.path import join

            def func():
                join = "this" + "joined with" + "this"
                return join
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Tuple assignment
        _ast = parse(
            """
            from os.path import join

            def func():
                a, b, join, d = 1, 2, 3, 4
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_annassign_to_imported_name(self, parse, capfd):
        _ast = parse(
            """
            import math

            math: str = "this is very bad"
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_augassign_to_imported_name(self, parse, capfd):
        _ast = parse(
            """
            from math import pi

            pi += 1.0   # that's not pi!
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_delete_of_imported_name(self, parse, capfd):
        # Single
        _ast = parse(
            """
            from math import pi

            def fn():
                del pi
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "attempt to delete imported name" in output
        assert _exit.called

        # Tuple
        _ast = parse(
            """
            from math import pi

            def fn():
                del a, b, pi
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "attempt to delete imported name" in output
        assert _exit.called

    def test_function_def(self, parse, capfd):
        # Root
        _ast = parse(
            """
            from os.path import join

            def join(*args):
                return "/".join(args)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Nested
        _ast = parse(
            """
            from os.path import join

            def wrapper():
                def join(*args):
                    return "/".join(args)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_async_function_def(self, parse, capfd):
        # Root
        _ast = parse(
            """
            from os.path import join

            async def join(*args):
                return "/".join(args)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_class_def(self, parse, capfd):
        _ast = parse(
            """
            from module.submodule import SomeClass

            class SomeClass:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_respect_rattr_ignore(self, parse):
        _ast = parse(
            """
            from math import func, async_func, SomeClass

            @rattr_ignore
            def func():
                pass

            @rattr_ignore
            async def async_func():
                pass

            @rattr_ignore
            class SomeClass:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        assert not _exit.called

    def test_respect_rattr_results(self, parse):
        _ast = parse(
            """
            from math import func, async_func, SomeClass

            @rattr_results
            def func():
                pass

            @rattr_results
            async def async_func():
                pass

            @rattr_results
            class SomeClass:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        assert not _exit.called

    def test_argument_name(self, parse, capfd):
        # Normal
        _ast = parse(
            """
            from math import pi

            def area(r, pi):
                return pi * r * r
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Nested
        _ast = parse(
            """
            from math import pi

            def wrapped():
                def area(r, pi):
                    return pi * r * r
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Async
        _ast = parse(
            """
            from math import pi

            async def area(r, pi):
                return pi * r * r
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_lambda(self, parse, capfd):
        _ast = parse(
            """
            from math import pi

            x = lambda pi: 3.14
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_for(self, parse, capfd):
        # For
        _ast = parse(
            """
            from math import pi

            for pi in pies:
                eat(pi)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Tuple
        _ast = parse(
            """
            from math import pi

            for a, pi in pies:
                eat(pi)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Async
        _ast = parse(
            """
            from math import pi

            async for pi in pies:
                eat(pi)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_with(self, parse, capfd):
        # With
        _ast = parse(
            """
            from math import pi

            with whatever as pi:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Multi
        _ast = parse(
            """
            from math import pi

            with whatever as we, whenever as pi:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Async
        _ast = parse(
            """
            from math import pi

            async with whatever as pi:
                pass
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_comprehension(self, parse, capfd):
        # List
        _ast = parse(
            """
            from math import pi

            l = [pi for pi in pies]
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Set
        _ast = parse(
            """
            from math import pi

            s = {pi for _, pi in pastries_and_pies}
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Generator expression
        _ast = parse(
            """
            from math import pi

            is_any = any(pi for pi in pies)
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

        # Dict
        _ast = parse(
            """
            from math import pi

            d = {pi: filling for pi, filling in pies}
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" in output
        assert _exit.called

    def test_do_not_fail_on_write_to_attribute(self, parse, capfd):
        # Direct
        _ast = parse(
            """
            from my_module import MyClass

            MyClass.attr = "value"
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" not in output
        assert _exit.called

        # Indirect
        _ast = parse(
            """
            from my_module import MyClass

            thing = MyClass.attr = "value"
            """
        )
        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(
                _ast, RootContext(_ast)
            )

        output, _ = capfd.readouterr()

        assert "redefinition of imported name" not in output
        assert _exit.called



"""Rattr error/logging functions."""

import ast
import sys
from enum import Enum
from os.path import expanduser
from typing import List, Optional, Tuple

from rattr import config

__ERROR = "{prefix}: {optional_file_info}{optional_line_info}{message}"
__FILE_INFO = "\033[1m{}: \033[0m"
__LINE_INFO = "\033[1mline {}:{}: \033[0m"


# --------------------------------------------------------------------------- #
# Rattr errors
# --------------------------------------------------------------------------- #


class Level(Enum):
    RATTR = "\033[34;1mrattr\033[0m"  # Blue
    INFO = "\033[33;1minfo\033[0m"  # Yellow / Orange
    WARNING = "\033[33;1mwarning\033[0m"  # Yellow / Orange
    ERROR = "\033[31;1merror\033[0m"  # Red
    FATAL = "\033[31;1mfatal\033[0m"  # Red


def rattr(
    message: str,
    culprit: Optional[ast.AST] = None,
    badness: int = 0,  # noqa
) -> None:
    """Log a message with the prefix "rattr", not for analyser errors."""
    __log(Level.RATTR, message, culprit)


def info(
    message: str,
    culprit: Optional[ast.AST] = None,
    badness: int = 0,
) -> None:
    """Log a low-priority warning and, if given, include culprit info."""
    __increment_badness(badness)

    if not config.show_warnings:
        return

    if not config.show_low_priority_warnings:
        return

    if not config.show_imports_warnings and config.current_file != config.file:
        return

    __log(Level.INFO, message, culprit)


def warning(
    message: str,
    culprit: Optional[ast.AST] = None,
    badness: int = 1,
) -> None:
    """Log a warning and, if given, include culprit line and file info."""
    __increment_badness(badness)

    if not config.show_warnings:
        return

    if not config.show_imports_warnings and config.current_file != config.file:
        return

    __log(Level.WARNING, message, culprit)


def error(
    message: str,
    culprit: Optional[ast.AST] = None,
    badness: int = 5,
) -> None:
    """Log an error and, if given, include culprit line and file info."""
    __increment_badness(badness)

    if config.strict and badness > 0:
        fatal(message, culprit)

    __log(Level.ERROR, message, culprit)


def fatal(
    message: str,
    culprit: Optional[ast.AST] = None,
    badness: int = 0,  # noqa
) -> None:
    """Log a fatal error and, if given, include culprit line and file info.

    NOTE
        A fatal error has no notion of badness as it will always cause an
        immediate EXIT_FAILURE, however, badness is provided in the function
        interface for consistency with the other errors.

        Regardless of the provided badness value, a badness of 0 will be used.

    """
    __increment_badness(0)

    __log(Level.FATAL, message, culprit)

    sys.exit(1)


def get_badness() -> int:
    """Return the badness value."""
    return config.file_badness + config.simplify_badness


def is_within_badness_threshold() -> bool:
    """Return `True` if the program is within the current badness threshold."""
    badness = get_badness()

    if config.strict:
        return badness <= 0

    # NOTE A threshold of 0 is equivalent to a threshold of ∞
    if config.threshold == 0:
        return True

    return badness <= config.threshold


# --------------------------------------------------------------------------- #
# Raisable errors
# --------------------------------------------------------------------------- #


class RattrUnsupportedError(Exception):
    """Language feature is unsupported by Rattr."""

    pass


class RattrUnaryOpInNameable(TypeError):
    """Unary operation found when resolving name."""

    pass


class RattrBinOpInNameable(TypeError):
    """Binary operation found when resolving name."""

    pass


class RattrConstantInNameable(TypeError):
    """Constant found when resolving name."""

    pass


class RattrLiteralInNameable(TypeError):
    """Literal found when resolving name."""

    pass


class RattrComprehensionInNameable(TypeError):
    """Comprehension found when resolving name."""

    pass


# --------------------------------------------------------------------------- #
# Error utils
# --------------------------------------------------------------------------- #


def get_file_and_line_info(culprit: Optional[ast.AST]) -> Tuple[str, str]:
    """Return the formatted line and line and file info as strings."""
    if culprit is None:
        return "", ""

    if config.current_file is not None and config.show_path:
        file_info = __FILE_INFO.format(format_path(config.current_file))
    else:
        file_info = ""

    line_info = __LINE_INFO.format(culprit.lineno, culprit.col_offset)

    return file_info, line_info


def split_path(path: str) -> List[str]:
    """Return the components of the path.

    >>> path == "/".join(split_path(path))
    True
    for all path

    >>> split_path("a/b/c")
    ["a", "b", "c"]

    >>> split_path("/a/b/c")
    ["", "a", "b", "c"]

    >>> split_path("~/a/b/c")
    ["~", "a", "b", "c"]

    """
    if path in ("", "/"):
        return [""]

    # if path.startswith("/"):
    #     return ["/"] + path[1:].split("/")

    if not path.startswith((".", "~", "/")):
        path = f"./{path}"

    return path.split("/")


def format_path(path: Optional[str]) -> Optional[str]:
    """Return the given path formatted in line with `config`."""
    if path is None:
        return None

    if not config.use_short_path:
        return path

    # Replace $HOME with "~"
    path = path.replace(expanduser("~"), "~")

    # Abbreviate long heirachies
    segments = split_path(path)
    if len(segments) > 5:
        path = "/".join([segments[0], "...", *segments[-3:]])

    return path


def __log(
    level: Level,
    message: str,
    culprit: Optional[ast.AST] = None,
) -> None:
    file_info, line_info = get_file_and_line_info(culprit)

    print(
        __ERROR.format(
            prefix=level.value,
            optional_file_info=file_info,
            optional_line_info=line_info,
            message=message,
        )
    )


def __increment_badness(badness: int) -> None:
    if isinstance(badness, int) and badness < 0:
        raise ValueError("'badness' must be positive integer")

    if config.current_file == config.file:
        config.file_badness += badness
    elif config.current_file is None:
        config.simplify_badness += badness
    else:
        config.import_badness += badness



"""Rattr class analyser.

Rattr is designed to be a function analyser and thus the class-based features
are rather limited.

"""

import ast
from typing import List

from rattr.analyser.base import NodeVisitor
from rattr.analyser.context import Context
from rattr.analyser.context.symbol import Class, Func, Name, Symbol
from rattr.analyser.function import FunctionAnalyser
from rattr.analyser.types import (
    AnyAssign,
    AnyFunctionDef,
    ClassIR,
    FunctionIR,
    ast_NamedExpr,
)
from rattr.analyser.util import (
    get_assignment_targets,
    get_fullname,
    get_function_def_args,
    has_annotation,
    unravel_names,
)


def is_method(node: ast.AST) -> bool:
    return isinstance(node, (ast.AsyncFunctionDef, ast.FunctionDef))


def get_statements(statements: ast.AST) -> List[AnyFunctionDef]:
    return list(filter(lambda stmt: not is_method(stmt), statements))


def get_methods(statements: ast.AST) -> List[ast.stmt]:
    return list(filter(lambda stmt: is_method(stmt), statements))


def get_initialiser(methods: List[AnyFunctionDef]) -> AnyFunctionDef:
    init = list(filter(lambda m: m.name == "__init__", methods))

    if len(init) == 0:
        return None

    return init[0]


def get_static_methods(methods: List[AnyFunctionDef]) -> AnyFunctionDef:
    return filter(lambda m: has_annotation("staticmethod", m), methods)


def get_base_names(cls: ast.ClassDef) -> List[str]:
    return list(map(lambda b: get_fullname(b, safe=True), cls.bases))


def is_enum(cls: ast.ClassDef) -> bool:
    # NOTE Purely heuristic, though it does allow for user defined Enum bases
    return any(n == "Enum" or n.endswith(".Enum") for n in get_base_names(cls))


class ClassAnalyser(NodeVisitor):
    """Walk a class's AST to determine the accessed attributes.


    TODO
        - Inheritance, super
        - Class methods
        - Methods from type
        - Nested classes

    """

    def __init__(self, _ast: ast.ClassDef, context: Context) -> None:
        """Set configuration and initialise IR."""
        if not isinstance(_ast, ast.ClassDef):
            raise TypeError("ClassAnalyser expects `_ast` to be a class")

        self._ast = _ast
        self.class_name = _ast.name

        self.class_ir: ClassIR = dict()

        # NOTE Managed by `new_context` contextmanager -- do not manually set
        self.context = context

    def analyse(self) -> ClassIR:
        """Entry point, return the IR produced from analysis.

        Algorithm:
            * Analyse non-method body -- i.e. class attributes, etc
                Add to context as `ClassName.attr`
                TODO Allow reference from class body context as just `attr`
            * Analyse __init__
                Add to context as Func: `ClassName`
                TODO deal with __new__, __new__ and __init__, etc
            * Analyse `@static_methods`
                Add to context as Func: `ClassName.method_name`
            * TODO Analyse `@class_methods`
            * TODO Analyse remaining methods as object methods

        """

        # NOTE
        #   Analyse within parent context, i.e. no new_context, as static
        #   methods, initialisers, should be present in the parent context
        #   just with transformed names

        statements = get_statements(self._ast.body)
        methods = get_methods(self._ast.body)

        # Visit statements
        for stmt in statements:
            self.visit(stmt)

        # Visit initialiser
        init = get_initialiser(methods)

        if init is not None:
            self.visit_initialiser(init)

        # NOTE TODO Handle with proper inheritance
        if init is None and is_enum(self._ast):
            self.visit_enum_initialiser()

        # Visit static methods
        for method in get_static_methods(methods):
            self.visit_static_method(method)

        return self.class_ir

    # ----------------------------------------------------------------------- #
    # Non-method statements
    # ----------------------------------------------------------------------- #

    def visit_initialiser(self, init: AnyFunctionDef) -> None:
        # NOTE
        #   Class already in context (as Class("ClassName"))
        #   Add initialiser as Func("ClassName.__init__")
        #   On resolving call with target Class("ClassName") look for
        #   Func("ClassName.__init__", ...)
        cls: Class = self.context.get(self.class_name)

        # Update the class symbol with __init__'s arguments
        # See: context/symbol.py::Class
        # See: context/context.py::RootContext.register_ClassDef
        cls.args, cls.vararg, cls.kwarg = get_function_def_args(init)

        self.class_ir[cls] = FunctionAnalyser(init, self.context).analyse()

    def visit_enum_initialiser(self) -> None:
        cls: Class = self.context.get(self.class_name)
        cls.args, cls.vararg, cls.kwarg = (
            [
                "self",
                "_id",
            ],
            None,
            None,
        )

        # NOTE Can't determine result at compile-time, thus give every option
        symbols = self.context.symbol_table.symbols()
        names: List[Symbol] = list(filter(lambda s: isinstance(s, Name), symbols))
        enum_values = filter(lambda n: n.name.startswith(f"{self.class_name}."), names)

        ir: FunctionIR = {
            "sets": set(),
            "gets": set(enum_values),
            "calls": set(),
            "dels": set(),
        }

        self.class_ir[cls] = ir

    def visit_static_method(self, method: AnyFunctionDef) -> None:
        qualified_name = f"{self.class_name}.{method.name}"
        fn = Func(qualified_name, *get_function_def_args(method))

        self.class_ir[fn] = FunctionAnalyser(method, self.context).analyse()

    # ----------------------------------------------------------------------- #
    # Non-method statements
    # ----------------------------------------------------------------------- #

    def visit_AnyAssign(self, node: AnyAssign) -> None:
        """Helper method for assignments.

        Visit ast.Assign(targets, value, type_comment)\\
        Visit ast.AnnAssign(target, annotation, value, simple)\\
        Visit ast.AugAssign(target, op, value)

        """
        targets = get_assignment_targets(node)
        names = [name for target in targets for name in unravel_names(target)]

        for name in names:
            name = Name(f"{self.class_name}.{name}", self.class_name)
            self.context.add(name)

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast_NamedExpr) -> None:
        self.visit_AnyAssign(node)



"""Rattr symbol table."""

from typing import KeysView, ValuesView

from rattr.analyser.context.symbol import Import, Symbol


class SymbolTable(dict):
    """A map from names to symbols w.r.t. some context."""

    def add(self, symbol: Symbol) -> None:
        """Add `symbol.name -> symbol` to the symbol table."""
        name = symbol.name

        # NOTE Prevent name clash with multiple starred imports
        if isinstance(symbol, Import) and symbol.name == "*":
            name = f"{symbol.qualified_name}.*"

        self[name] = symbol

    def remove(self, name: str) -> None:
        """Remove the symbol with the given name, it it exists."""
        if name not in self.names():
            return

        del self[name]

    def names(self) -> KeysView[str]:
        """Return an iterator over the symbol table names."""
        return super().keys()

    def symbols(self) -> ValuesView[Symbol]:
        """Return an iterator over the symbol table symbols."""
        return super().values()



from argparse import ArgumentError, Namespace

import pytest

from rattr.cli.parser import parse_arguments


class TestParser:
    def test_toml_dict_with_illegal_fields(
        self, toml_dict_with_illegal_fields, illegal_field_name, sys_args1
    ):
        args = parse_arguments(
            sys_args=sys_args1,
            project_toml_cfg=toml_dict_with_illegal_fields,
            exit_on_error=False,
        )
        # illegal field should be left out from parsed args
        assert hasattr(args, illegal_field_name) is False

    def test_toml_dict_with_bad_field_types(
        self, toml_dicts_with_bad_field_types_and_errors
    ):
        for toml_dict, error_message in toml_dicts_with_bad_field_types_and_errors:
            with pytest.raises(ArgumentError) as exc_info:
                parse_arguments(project_toml_cfg=toml_dict, exit_on_error=False)
            assert str(exc_info.value) == error_message

    def test_toml_dicts_with_mutex_arg_violation(
        self, toml_dicts_with_mutex_arg_violation_and_errors
    ):
        for toml_dict, error_message in toml_dicts_with_mutex_arg_violation_and_errors:
            with pytest.raises(ArgumentError) as exc_info:
                parse_arguments(project_toml_cfg=toml_dict, exit_on_error=False)
            assert str(exc_info.value) == error_message

    def test_toml_dict_without_file_from_cli_call(self, correct_toml_dict1):
        with pytest.raises(ArgumentError) as exc_info:
            parse_arguments(
                # empty 'sys_args' imply no file specified in cli call
                sys_args=[],
                project_toml_cfg=correct_toml_dict1,
                exit_on_error=False,
            )

        error_message = "rattr: error: the following arguments are required: <file>\n"
        assert str(exc_info.value) == error_message

    def test_strict_toml_dict_with_file_from_cli_call(
        self, sys_args1, correct_toml_dict1
    ):
        args = parse_arguments(
            sys_args=sys_args1[1:],
            project_toml_cfg=correct_toml_dict1,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=3,
            exclude_import=["a\\.a\\.a", "b\\.b.*", "c\\.c\\.c\\.c", "d\\d\\.d.*"],
            exclude=["a_.*", "b_.*", "c_.*"],
            show_warnings="all",
            show_path="short",
            strict=True,
            permissive=False,
            show_ir=False,
            show_results=True,
            show_stats=False,
            silent=False,
            cache="cache.json",
            threshold=0,
            filter_string="",
            file="rattr/cli/parser.py",
        )

        assert args == exp_args

    def test_permissive_toml_dict_with_file_from_cli_call(
        self, sys_args1, correct_toml_dict2
    ):
        args = parse_arguments(
            sys_args=sys_args1[1:],
            project_toml_cfg=correct_toml_dict2,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=3,
            exclude_import=["a\\.a\\.a", "b\\.b.*", "c\\.c\\.c\\.c", "d\\d\\.d.*"],
            exclude=["a_.*", "b_.*", "c_.*"],
            show_warnings="all",
            show_path="full",
            strict=False,
            permissive=True,
            show_ir=False,
            show_results=True,
            show_stats=False,
            silent=False,
            cache="cache.json",
            threshold=1,
            filter_string="",
            file="rattr/cli/parser.py",
        )

        assert args == exp_args

    def test_toml_dict_with_overwrites_from_cli_call(
        self, sys_args2, correct_toml_dict1
    ):
        args = parse_arguments(
            sys_args=sys_args2[1:],
            project_toml_cfg=correct_toml_dict1,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=2,
            exclude_import=["a\\.a\\.a", "b\\.b.*", "c\\.c\\.c\\.c", "d\\d\\.d.*"],
            exclude=["a_.*", "b_.*", "c_.*"],
            show_warnings="file",
            show_path="short",
            strict=True,
            permissive=False,
            show_ir=False,
            show_results=True,
            show_stats=False,
            silent=False,
            cache="cache.json",
            threshold=0,
            filter_string="",
            file="rattr/cli/argparse.py",
        )

        assert args == exp_args

    def test_toml_dict_with_append_and_overwrites_from_cli_call(
        self, sys_args3, correct_toml_dict1
    ):
        args = parse_arguments(
            sys_args=sys_args3[1:],
            project_toml_cfg=correct_toml_dict1,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=2,
            exclude_import=[
                "a\\.a\\.a",
                "b\\.b.*",
                "c\\.c\\.c\\.c",
                "d\\d\\.d.*",
                "*exclude-import*",
            ],
            exclude=["a_.*", "b_.*", "c_.*", "*exclude-pattern*"],
            show_warnings="file",
            show_path="short",
            strict=True,
            permissive=False,
            show_ir=False,
            show_results=False,
            show_stats=False,
            silent=True,
            cache="cache.json",
            threshold=0,
            filter_string="",
            file="rattr/cli/argparse.py",
        )

        assert args == exp_args

    def test_toml_dict_with_overwrites_and_mutex_arg_switch(
        self, sys_args2, correct_toml_dict2
    ):
        args = parse_arguments(
            sys_args=sys_args2[1:],
            project_toml_cfg=correct_toml_dict2,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=2,
            exclude_import=["a\\.a\\.a", "b\\.b.*", "c\\.c\\.c\\.c", "d\\d\\.d.*"],
            exclude=["a_.*", "b_.*", "c_.*"],
            show_warnings="file",
            show_path="short",
            strict=True,
            permissive=False,
            show_ir=False,
            show_results=True,
            show_stats=False,
            silent=False,
            cache="cache.json",
            threshold=0,
            filter_string="",
            file="rattr/cli/argparse.py",
        )

        assert args == exp_args

    def test_toml_dict_with_mutex_violation_cli_overwrite(
        self, sys_args4, correct_toml_dict3
    ):

        with pytest.raises(ArgumentError) as exc_info:
            parse_arguments(
                sys_args=sys_args4[1:],
                project_toml_cfg=correct_toml_dict3,
                exit_on_error=False,
            )

        error_message = "-irs ('--show-ir', '--show-results', '--show-stats') and -S ('--silent') are mutually exclusive"  # noqa: E501
        assert str(exc_info.value) == error_message

    def test_empty_toml_dict_with_only_cli_call(self, sys_args4):
        args = parse_arguments(
            # empty 'sys_args' imply no file specified in cli call
            sys_args=sys_args4[1:],
            project_toml_cfg={},
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=1,
            exclude_import=[],
            exclude=[],
            show_warnings="all",
            show_path="short",
            strict=False,
            permissive=True,
            show_ir=False,
            show_results=False,
            show_stats=False,
            silent=True,
            cache="",
            filter_string="",
            file="rattr/cli/parser.py",
            threshold=0,
        )

        assert args == exp_args

    def test_permissive_cli_call_overwrite(self, sys_args5, correct_toml_dict2):
        args = parse_arguments(
            sys_args=sys_args5[1:],
            project_toml_cfg=correct_toml_dict2,
            exit_on_error=False,
        )

        exp_args = Namespace(
            follow_imports=3,
            exclude_import=["a\\.a\\.a", "b\\.b.*", "c\\.c\\.c\\.c", "d\\d\\.d.*"],
            exclude=["a_.*", "b_.*", "c_.*"],
            show_warnings="all",
            show_path="full",
            strict=False,
            permissive=True,
            show_ir=False,
            show_results=True,
            show_stats=False,
            silent=False,
            cache="cache.json",
            threshold=3,
            filter_string="",
            file="rattr/cli/parser.py",
        )

        assert args == exp_args



from abc import ABC, abstractstaticmethod
from argparse import ArgumentError, Namespace, RawTextHelpFormatter
from os.path import isfile, splitext
from typing import Any, Dict, List, Optional, Tuple

from tomli import TOMLDecodeError

from rattr import _version, error
from rattr.cli.argparse import _ArgumentParser as ArgumentParser
from rattr.cli.toml_parser import find_project_toml, load_cfg_from_project_toml
from rattr.cli.util import multi_paragraph_wrap


def translate_toml_cfg_to_sys_args(toml_cfg: Dict[str, Any]) -> List[str]:
    """
    Function translates pyproject.toml config dict into a sys.argv style list
    of string args (like: ['--permissive', '0', '--show-warnings', 'all', '--silent'])
    and then this is passed to the argument parses which then validates these args.
    """
    toml_sys_args = []
    for k, v in toml_cfg.items():
        arg_name = f"-{k}" if len(str(k)) == 1 else f"--{k}"
        if isinstance(v, bool):
            toml_sys_args += [arg_name] if v else []
        elif isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
            toml_sys_args += [arg_name, f"{v}"]
        elif isinstance(v, list):
            for arg_value in v:
                toml_sys_args += [arg_name, f"{arg_value}"]
    return toml_sys_args


def validate_toml_cfg_arg_types(
    toml_cfg: Dict[str, Any],
    arg_group_parsers: Tuple[ABC],  # ArgumentGroupParser classes
) -> Dict[str, Any]:
    """
    Function takes the 'toml_cfg' dictionary and checks each field against
    its corresponding 'ArgumentGroupParser' class. Each ArgumentGroupParsers
    class have a 'TOML_ARG_NAME_ARG_TYPE_MAP' dict property which is a map of
    'toml_field_name' -> 'toml_field_type'.
    """
    for parser in arg_group_parsers:
        arg_name_type_map = parser.TOML_ARG_NAME_ARG_TYPE_MAP
        for arg_name, expected_arg_type in arg_name_type_map.items():
            try:
                toml_arg_val = toml_cfg[arg_name]
                if not isinstance(toml_arg_val, expected_arg_type):
                    message = (
                        f"Error parsing pyproject.toml. Arg: "
                        f"'{arg_name}' is of wrong type: "
                        f"'{type(toml_arg_val).__name__}'. "
                        f"Expected type: '{expected_arg_type.__name__}'."
                    )
                    raise ArgumentError(None, message)
            except KeyError:
                pass
    return toml_cfg


def parse_arguments(
    sys_args: Optional[List[str]] = None,
    project_toml_cfg: Optional[Dict[str, Any]] = None,
    exit_on_error: bool = True,
) -> Namespace:
    """Return the parsed and validated CLI arguments.

    The parser is constructed and validated using the registered
    `ArgumentGroupParser`s.

    Each `ArgumentGroupParser::register` adds an
    `argparse.ArgumentParser::add_argument_group` to the parser and returns the
    parser.

    Each `ArgumentGroupParser::validate` validates the current state of the
    arguments, modifies them if need be, and returns the resulting arguments.
    If there is invalidity within the arguments, then the method should raise
    and `argparse.ArgumentError` with an appropriate message.

    On error the program usage and the error message will be shown before
    exiting.

    """

    cli_parser: ArgumentParser = ArgumentParser(
        prog="rattr",
        description=multi_paragraph_wrap(
            """\
            Parse a given Python 3 file to find the attributes used in each
            function.

            Given the regular expression <filter-string>, the results will only
            include functions, classes, or methods whose name fully matches
            regular expression.

            Uses Python's regular expression syntax.
            """
        ),
        epilog=multi_paragraph_wrap(
            """\
            Made with ❤️ by Suade Labs.
            """
        ),
        formatter_class=RawTextHelpFormatter,
        exit_on_error=exit_on_error,
    )
    # don't exit on error when .toml parser fails since we want
    # to construct our own error message saying error is from .toml file
    toml_parser: ArgumentParser = ArgumentParser(exit_on_error=False)

    # Version
    version_group = cli_parser.add_argument_group()
    version_group.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {_version.version}",
    )

    # TODO Allow user to add to this (will need config to respect this)

    # only these should be configurable in .toml file (filter and file )
    # should come from cli args
    TOML_ARG_GROUP_PARSERS = (
        FollowImports,
        ExcludeImports,
        ExcludePatterns,
        ShowWarnings,
        ShowPath,
        StrictOrPermissive,
        Output,
        Cache,
    )

    CLI_ARG_GROUP_PARSERS = TOML_ARG_GROUP_PARSERS + (FilterString, File)

    for argument_group_parser in CLI_ARG_GROUP_PARSERS:
        cli_parser = argument_group_parser.register(cli_parser)
        if argument_group_parser in TOML_ARG_GROUP_PARSERS:
            toml_parser = argument_group_parser.register(toml_parser)

    # TODO
    #   One moved to Python 3.9+ add `exit_on_error=False` to the
    #   `ArgumentParser` constructor and catch the error the same as below --
    #   this will give consistent behaviour between parser and validator errors

    # extracting .toml expected field lists from 'TOML_ARG_GROUP_PARSERS'
    toml_expected_fields: Tuple[str] = ()
    for arg_group_parser in TOML_ARG_GROUP_PARSERS:
        toml_expected_fields += tuple(
            arg_group_parser.TOML_ARG_NAME_ARG_TYPE_MAP.keys()
        )

    if not project_toml_cfg:
        try:
            toml_cfg_path = find_project_toml()
            project_toml_cfg = load_cfg_from_project_toml(toml_cfg_path=toml_cfg_path)
        except TOMLDecodeError as e:
            if exit_on_error:
                error.fatal(
                    f"Error decoding pyproject.toml file: {toml_cfg_path}. Error: {e}."
                )
            raise e

    # only keep expected fields in toml cfg dict
    project_toml_cfg = {
        k: v for k, v in project_toml_cfg.items() if k in toml_expected_fields
    }

    try:
        project_toml_cfg = validate_toml_cfg_arg_types(
            toml_cfg=project_toml_cfg,
            arg_group_parsers=TOML_ARG_GROUP_PARSERS,
        )
    except ArgumentError as e:
        if exit_on_error:
            error.fatal(e.message)
        raise e

    # construct 'sys.argv' style list of args from 'project_toml_cfg' dict
    toml_cfg_arg_list = translate_toml_cfg_to_sys_args(toml_cfg=project_toml_cfg)

    # namespace for collecting args from .toml file and sys.argv/cli args
    arguments = Namespace()

    # collect .toml args into 'Namespace' object first
    try:
        # parse args with toml_parser (which validates args at the same time),
        # this way we don't need to create a custom validation module
        # for the .toml config for cases such as mutex arg group validation etc...
        arguments = toml_parser.parse_args(args=toml_cfg_arg_list, namespace=arguments)
    except ArgumentError as e:
        if exit_on_error:
            message = (
                f"Error parsing pyproject.toml. Arg: "
                f"'{e.argument_name}', Error: {e.message}."
            )
            error.fatal(message)
        raise e

    # validate .toml args against validators
    for argument_group_parser in TOML_ARG_GROUP_PARSERS:
        try:
            if argument_group_parser is not Output:
                arguments = argument_group_parser.validate(toml_parser, arguments)
                continue
            # turn of show-results defaulting behaviour from toml file
            arguments = argument_group_parser.validate(cli_parser, arguments, toml=True)
        except ArgumentError as e:
            if exit_on_error:
                message = f"Error parsing pyproject.toml. Error: {e.message}."
                error.fatal(message)
            raise e

    # then parse cli args and overwrite overlapping toml args
    arguments = cli_parser.parse_args(
        # use empty arg Namespace (without set defaults from toml arg parsing)
        # when project_toml_cfg is empty
        namespace=arguments if project_toml_cfg else Namespace(),
        args=sys_args,
    )

    # validate args once again
    for argument_group_parser in CLI_ARG_GROUP_PARSERS:
        try:
            arguments = argument_group_parser.validate(cli_parser, arguments)
        except ArgumentError as e:
            if exit_on_error:
                error.rattr(cli_parser.format_usage())
                error.fatal(str(e))
            raise e

    return arguments


class ArgumentGroupParser(ABC):

    TOML_ARG_NAME_ARG_TYPE_MAP = {}

    @abstractstaticmethod
    def register(parser: ArgumentParser) -> ArgumentParser:
        return parser

    @abstractstaticmethod
    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        return arguments


class FollowImports(ArgumentGroupParser):

    ARG_LONG_NAME = "follow-imports"

    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: int}

    def register(parser: ArgumentParser) -> ArgumentParser:
        follow_imports_group = parser.add_argument_group()
        follow_imports_group.add_argument(
            "-f",
            f"--{FollowImports.ARG_LONG_NAME}",
            default=1,
            type=int,
            choices=[0, 1, 2, 3],
            help=multi_paragraph_wrap(
                """\
                >follow imports level meanings:
                >    0 - do not follow imports
                >    1 - follow imports to local modules \033[1m(default)\033[0m
                >    2 - follow imports to local and pip installed modules
                >    3 - follow imports to local, pip installed, and stdlib modules
                >NB: following stdlib imports when using CPython will cause issues
                >TOML example: follow-imports=1
                """
            ),
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        if not arguments.follow_imports:
            error.rattr("follow imports not set, results may be incomplete")

        return arguments


class ExcludeImports(ArgumentGroupParser):

    ARG_LONG_NAME = "exclude-import"

    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: list}

    def register(parser: ArgumentParser) -> ArgumentParser:
        exclude_imports_group = parser.add_argument_group()
        exclude_imports_group.add_argument(
            "-F",
            f"--{ExcludeImports.ARG_LONG_NAME}",
            action="append",
            type=str,
            help=multi_paragraph_wrap(
                """\
                >do not follow imports to modules matching the given pattern,
                >regardless of the level of \033[1m-f\033[0m
                >TOML example: exclude-imports=['a', 'b']
                """
            ),
            metavar="PATTERN",
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        if arguments.exclude_import is None:
            arguments.exclude_import = list()

        return arguments


class ExcludePatterns(ArgumentGroupParser):

    ARG_LONG_NAME = "exclude"

    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: list}

    def register(parser: ArgumentParser) -> ArgumentParser:
        exclude_patterns_group = parser.add_argument_group()
        exclude_patterns_group.add_argument(
            "-x",
            f"--{ExcludePatterns.ARG_LONG_NAME}",
            action="append",
            type=str,
            help=multi_paragraph_wrap(
                """\
                >exclude functions and classes matching the
                >given regular expression from being analysed
                >TOML example: exclude=['a', 'b']
                """
            ),
            metavar="PATTERN",
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        if arguments.exclude is None:
            arguments.exclude = list()

        return arguments


class ShowWarnings(ArgumentGroupParser):

    ARG_LONG_NAME = "show-warnings"

    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: str}

    def register(parser: ArgumentParser) -> ArgumentParser:
        # TODO rename: all -> default, ALL -> all
        show_warnings_group = parser.add_argument_group()
        show_warnings_group.add_argument(
            "-w",
            f"--{ShowWarnings.ARG_LONG_NAME}",
            default="all",
            type=str,
            choices=["none", "file", "all", "ALL"],
            help=multi_paragraph_wrap(
                """\
                >show warnings level meaning:
                >    none - do not show warnings
                >    file - show warnings for <file>
                >    all  - show warnings for all files \033[1m(default)\033[0m
                >    All  - show warnings for all files, including low-priority
                >NB: errors and fatal errors are always shown
                >TOML example: show-warnings='all'
                """
            ),
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        return arguments


class ShowPath(ArgumentGroupParser):

    ARG_LONG_NAME = "show-path"

    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: str}

    def register(parser: ArgumentParser) -> ArgumentParser:
        show_path_group = parser.add_argument_group()
        show_path_group.add_argument(
            "-p",
            f"--{ShowPath.ARG_LONG_NAME}",
            default="short",
            type=str,
            choices=["none", "short", "full"],
            help=multi_paragraph_wrap(
                """\
                >show path level meaning:
                >    none  - do not show the file path in errors/warnings
                >    short - show an abbreviated path \033[1m(default)\033[0m
                >    full  - show the full path
                >E.g.: "/home/user/very/deep/dir/path/file" becomes "~/.../dir/path/file"  # noqa
                >TOML example: show-path='short'
                """
            ),
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        return arguments


class StrictOrPermissive(ArgumentGroupParser):

    STRICT_ARG_LONG_NAME = "strict"
    PERMISSIVE_ARG_LONG_NAME = "permissive"

    TOML_ARG_NAME_ARG_TYPE_MAP = {
        STRICT_ARG_LONG_NAME: bool,
        PERMISSIVE_ARG_LONG_NAME: int,
    }

    def register(parser: ArgumentParser) -> ArgumentParser:
        strict_or_permissive_group = parser.add_argument_group()
        strict_or_permissive_mutex_group = (
            strict_or_permissive_group.add_mutually_exclusive_group()
        )

        strict_or_permissive_mutex_group.add_argument(
            f"--{StrictOrPermissive.STRICT_ARG_LONG_NAME}",
            action="store_true",
            help=multi_paragraph_wrap(
                """\
                >run rattr in strict mode, i.e. fail on any error
                >TOML example: strict=true
                """
            ),
        )
        strict_or_permissive_mutex_group.add_argument(
            f"--{StrictOrPermissive.PERMISSIVE_ARG_LONG_NAME}",
            default=0,
            type=int,
            help=multi_paragraph_wrap(
                """\
                run rattr in permissive mode, with the given badness threshold
                (when threshold is zero or omitted, it is taken as infinite)
                \033[1m(default: --permissive 0 when group omitted)\033[0m

                >typical badness values:
                >    +0 - info
                >    +1 - warning
                >    +5 - error
                >    +∞ - fatal

                >NB: badness is only contributed to by the target <file> and
                >by the simplification stage (e.g. resolving function calls, etc)
                >TOML example: permissive=10
                """
            ),
            metavar="THRESHOLD",
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:

        if not arguments.strict:
            # only set threshold and permissive
            # when permissive is freshly set to an int
            if isinstance(arguments.permissive, int):
                arguments.threshold = arguments.permissive or 0
                arguments.permissive = True
        else:
            arguments.threshold = 0
            arguments.permissive = False

        if arguments.threshold < 0:
            raise ArgumentError(None, "--permissive THRESHOLD must be positive")

        return arguments


class Output(ArgumentGroupParser):

    SHOW_IR_ARG_LONG_NAME = "show-ir"
    SHOW_RESULTS_ARG_LONG_NAME = "show-results"
    SHOW_STATS_ARG_LONG_NAME = "show-stats"
    SILENT_ARG_LONG_NAME = "silent"

    TOML_ARG_NAME_ARG_TYPE_MAP = {
        SHOW_IR_ARG_LONG_NAME: bool,
        SHOW_RESULTS_ARG_LONG_NAME: bool,
        SHOW_STATS_ARG_LONG_NAME: bool,
        SILENT_ARG_LONG_NAME: bool,
    }

    def register(parser: ArgumentParser) -> ArgumentParser:
        output_group = parser.add_argument_group()
        output_group.add_argument(
            "-i",
            f"--{Output.SHOW_IR_ARG_LONG_NAME}",
            action="store_true",
            help=multi_paragraph_wrap(
                """\
                >show the IR for the file and imports
                >TOML example: show-ir=true\n
                """
            ),
        )
        output_group.add_argument(
            "-r",
            f"--{Output.SHOW_RESULTS_ARG_LONG_NAME}",
            action="store_true",
            help=multi_paragraph_wrap(
                """\
                >show the results of analysis
                >TOML example: show-results=true\n
                """
            ),
        )
        output_group.add_argument(
            "-s",
            f"--{Output.SHOW_STATS_ARG_LONG_NAME}",
            action="store_true",
            help=multi_paragraph_wrap(
                """\
                >show stats Rattr statisitics
                >TOML example: show-stats=true\n
                """
            ),
        )
        output_group.add_argument(
            "-S",
            f"--{Output.SILENT_ARG_LONG_NAME}",
            action="store_true",
            help=multi_paragraph_wrap(
                """\
                >show only errors and warnings
                >TOML example: silent=true
                """
            ),
        )

        return parser

    def validate(
        parser: ArgumentParser, arguments: Namespace, toml: bool = False
    ) -> Namespace:
        has_output = any(
            (arguments.show_ir, arguments.show_results, arguments.show_stats)
        )

        # Output group mutual exclusion
        # [ [-irs] | -S ]
        if has_output and arguments.silent:
            message = (
                f"-irs ('--{Output.SHOW_IR_ARG_LONG_NAME}', "
                f"'--{Output.SHOW_RESULTS_ARG_LONG_NAME}', "
                f"'--{Output.SHOW_STATS_ARG_LONG_NAME}') "
                f"and -S ('--{Output.SILENT_ARG_LONG_NAME}') are mutually exclusive"
            )
            raise ArgumentError(None, message)

        # don't default to show-results = True when
        # output isn't configured in .toml since
        # if later cli parser brings silent mode (-S)
        # there will be a mutex validation error
        if not has_output and not arguments.silent and not toml:
            arguments.show_results = True

        return arguments


class FilterString(ArgumentGroupParser):
    def register(parser: ArgumentParser) -> ArgumentParser:
        filter_string_group = parser.add_argument_group()
        filter_string_group.add_argument(
            "filter_string",
            nargs="?",
            default="",
            type=str,
            help=multi_paragraph_wrap(
                """\
                >filter the output to functions matching the given regular expression
                >can't be specified in .toml
                """
            ),
            metavar="<filter-string>",
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        if arguments.filter_string == ".*":
            arguments.filter_string = ""

        return arguments


class File(ArgumentGroupParser):
    def register(parser: ArgumentParser) -> ArgumentParser:
        file_group = parser.add_argument_group()
        file_group.add_argument(
            "file",
            nargs=1,
            type=str,
            help=multi_paragraph_wrap(
                """\
                >the Python source file to analyse
                >can't be specified in .toml
                """
            ),
            metavar="<file>",
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        if len(arguments.file) != 1:
            raise ArgumentError(None, "expects exactly one file")

        arguments.file = arguments.file[0]
        _, ext = splitext(arguments.file)

        if not isfile(arguments.file):
            raise ArgumentError(None, f"file '{arguments.file}' not found")

        if ext != ".py":
            error.rattr(f"expects extension '.py', got '{ext}'")

        return arguments


class Cache(ArgumentGroupParser):

    ARG_LONG_NAME = "cache"
    TOML_ARG_NAME_ARG_TYPE_MAP = {ARG_LONG_NAME: str}

    def register(parser: ArgumentParser) -> ArgumentParser:
        cache_group = parser.add_argument_group()
        cache_group.add_argument(
            f"--{Cache.ARG_LONG_NAME}",
            default="",
            type=str,
            help=multi_paragraph_wrap(
                """\
                >the file to cache the results to, if successful
                >TOML example: cache='cache.json'
                """
            ),
        )

        return parser

    def validate(parser: ArgumentParser, arguments: Namespace) -> Namespace:
        f = arguments.cache
        _, ext = splitext(f)

        if isfile(f):
            error.rattr(f"'{f}' already exists and will be overwritten on success")

        if f != "" and ext != ".json":
            error.rattr(f"cache expects extension '.json', got '{ext}'")

        return arguments



import pytest

from rattr.analyser.context import Builtin, Call, Func, Name, RootContext
from rattr.analyser.file import FileAnalyser
from rattr.plugins import plugins
from rattr.plugins.analysers.collections import DefaultDictAnalyser


class TestCustomCollectionsAnalysers:
    @pytest.fixture(autouse=True)
    def apply_plugins(self):
        plugins.register(DefaultDictAnalyser())

    def test_defaultdict_no_factory(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                d = defaultdict()
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": set(),
                "sets": {
                    Name("d"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_defaultdict_named_factory(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                d = defaultdict(factory)

            def factory():
                return hello.results
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        a_func = Func("a_func", ["arg"], None, None)
        factory = Func("factory", [], None, None)
        expected = {
            a_func: {
                "gets": set(),
                "sets": {
                    Name("d"),
                },
                "dels": set(),
                "calls": {Call("factory()", [], {}, target=factory)},
            },
            factory: {
                "gets": {
                    Name("hello.results", "hello"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

    def test_defaultdict_lambda_factory(self, parse):
        # Lambda to literal
        _ast = parse(
            """
            def a_func(arg):
                d = defaultdict(lambda: 0)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": set(),
                "sets": {
                    Name("d"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        print(expected)
        assert results == expected

        # Lambda to attr
        _ast = parse(
            """
            def a_func(arg):
                d = defaultdict(lambda: arg.attr)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": {Name("arg.attr", "arg")},
                "sets": {
                    Name("d"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        print(expected)
        assert results == expected

    def test_defaultdict_nested_factory(self, parse):
        _ast = parse(
            """
            def a_func(arg):
                d = defaultdict(defaultdict(list))
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        list_builtin = Builtin("list", has_affect=False)
        expected = {
            Func("a_func", ["arg"], None, None): {
                "gets": set(),
                "sets": {
                    Name("d"),
                },
                "dels": set(),
                "calls": {Call("list()", [], {}, target=list_builtin)},
            }
        }

        assert results == expected



from rattr.error.error import (  # noqa
    RattrBinOpInNameable,
    RattrComprehensionInNameable,
    RattrConstantInNameable,
    RattrLiteralInNameable,
    RattrUnaryOpInNameable,
    RattrUnsupportedError,
    error,
    fatal,
    format_path,
    get_badness,
    get_file_and_line_info,
    info,
    is_within_badness_threshold,
    rattr,
    split_path,
    warning,
)



"""CustomFunctionAnalyser for Python's collections."""

import ast

from rattr.analyser.base import CustomFunctionAnalyser
from rattr.analyser.context import Context
from rattr.analyser.types import FuncOrAsyncFunc, FunctionIR


class DefaultDictAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "defaultdict"

    @property
    def qualified_name(self) -> str:
        return "collections.defaultdict"

    def on_def(self, name: str, node: FuncOrAsyncFunc, ctx: Context) -> FunctionIR:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIR:
        # HACK Avoid circular import
        from rattr.analyser.function import FunctionAnalyser

        if len(node.args) == 0:
            return super().on_call(name, node, ctx)

        default_factory = node.args[0]

        # Three cases:
        #   1. A named callable         ->  Add call to name
        #   2. A lambda                 ->  Fold-in results of body
        #   3. Non-lambda expression    ->  Visit

        fn_analyser = FunctionAnalyser(ast.FunctionDef(), ctx)

        if isinstance(default_factory, (ast.Name, ast.Attribute)):
            target = ast.Call(
                func=default_factory,
                args=[],
                keywords=[],
            )

            # Populate ast.AST fields
            target._fields = tuple()
            target.lineno = default_factory.lineno
            target.col_offset = default_factory.col_offset
        elif isinstance(default_factory, ast.Lambda):
            target = default_factory.body
        else:
            target = default_factory

        fn_analyser.visit(target)

        return fn_analyser.func_ir



from pathlib import Path
from typing import Any, Dict, Optional

try:
    # python >= 3.11
    import tomllib
except ImportError:
    # python < 3.11
    import tomli as tomllib


def load_cfg_from_project_toml(toml_cfg_path: Optional[Path]) -> Dict[str, Any]:
    """
    Function finds project toml and parses it into a toml config dictionary
    while only keeping expected fields (if expected fields is empty then
    keep all fields).
    """
    toml_cfg_path = find_project_toml() if not toml_cfg_path else toml_cfg_path
    if toml_cfg_path:
        return parse_project_toml(config_path=toml_cfg_path)
    return {}


def find_project_root() -> Optional[Path]:
    """
    Function finds project root by starting search from cwd and
    then exploring parent directories until it finds
    pyproject.toml or .git or .hg file.
    """
    srcs = [str(Path.cwd().resolve())]
    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]
    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )
    for directory in (common_base, *common_base.parents):
        if (
            (directory / ".git").exists()
            or (directory / ".hg").is_dir()
            or (directory / "pyproject.toml").is_file()
        ):
            return directory
    return None


def find_project_toml() -> Optional[str]:
    """
    Function finds pyproject.toml file by finding project root
    first then looking pyproject.toml there.
    """
    path_project_root = find_project_root()
    if not path_project_root:
        return None
    path_pyproject_toml = path_project_root / "pyproject.toml"
    if path_pyproject_toml.is_file():
        return str(path_pyproject_toml)
    return None


def parse_project_toml(config_path: str) -> Dict[str, Any]:
    """
    Function parses pyproject.toml file into a cfg dictionary
    while only keeping expected fields (if expected fields is empty
    then keep all fields).
    """
    with open(config_path, "rb") as f:
        pyproject_toml_cfg = tomllib.load(f)
    cfg, cleaned_cfg = pyproject_toml_cfg.get("tool", {}).get("rattr", {}), {}
    for k, v in cfg.items():
        if k.startswith("--"):
            k = k[2:]
        elif k.startswith("-"):
            k = k[1:]
        cleaned_cfg[k] = v
    return cleaned_cfg



"""Rattr analyser util functions."""

import ast
import builtins
import hashlib
import json
import re
import sys
from contextlib import contextmanager
from copy import deepcopy
from importlib.util import find_spec
from itertools import accumulate, chain, filterfalse
from os.path import isfile
from string import ascii_lowercase
from time import perf_counter
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from isort.api import place_module

from rattr import config, error
from rattr.analyser.context.symbol import get_possible_module_names  # noqa
from rattr.analyser.context.symbol import (
    Class,
    Import,
    Name,
    Symbol,
    parse_call,
    parse_name,
)
from rattr.analyser.types import (
    AnyAssign,
    AnyFunctionDef,
    AstDef,
    Comprehension,
    Constant,
    FileResults,
    FuncOrAsyncFunc,
    FunctionIR,
    Literal,
    Nameable,
    StrictlyNameable,
    ast_NamedExpr,
)

# The prefix given to local constants, literals, etc to produce a name
# E.g. "hi" -> get_basename_fullname_pair(.) = "@Str"
# This must be a character that is not legal for standard Python _identifiers
# or else code elsewhere may break
LOCAL_VALUE_PREFIX = "@"


# RegEx patterns for blacklisted modules
MODULE_BLACKLIST_PATTERNS = {
    "(rattr|rattr\\..*)",
    "(package.rattr|packages.rattr\\..*)",
}

# Python builtins which are not functions, classes, etc.
PYTHON_LITERAL_BUILTINS = {"None", "True", "False", "Ellipsis"}

# Python builtins which end in "attr", these dynamically access attributes on a
# given object
# NOTE
#   These must be handled explicitly by `FunctionAnalyser` or anyother
#   `ast.NodeVisitor` implementation that records results/context and sees them
PYTHON_ATTR_BUILTINS = {"delattr", "getattr", "hasattr", "setattr"}

# All Python builtins that are present in the `RootContext`
PYTHON_BUILTINS = set(filter(lambda b: b not in PYTHON_LITERAL_BUILTINS, dir(builtins)))


def get_basename_fullname_pair(
    node: Nameable,
    safe: bool = False,
) -> Tuple[str, str]:
    """Return the node's tuple of the basename and the fullname.

    If node (and it's children, recursively) are not StrictlyNameable, then a
    true name cannot be produced. If safe=True, then an attempt is made:

    #### All StrictlyNameable

        >>> get_basename_fullname_pair(
        ...     ast.parse("a.b[0].attr").body[0].value,
        ...     safe=True # or False
        ... )
        "a.b[].attr"

    #### Not all-StrictlyNameable, unsafe

        >>> get_basename_fullname_pair(
        ...     ast.parse("(a, b)[0].attr").body[0].value,
        ...     safe=False
        ... )
        error.RattrLiteralInNameable: ...

    #### Not all-StrictlyNameable, safe

        >>> get_basename_fullname_pair(
        ...     ast.parse("(a, b)[0].attr").body[0].value,
        ...     safe=True
        ... )
        "@Tuple[].attr"

    For the third example, the fact that `a.attr` is accessed is lost!

    """
    # Base case
    # ast.Name ⊂ StrictlyNameable
    if isinstance(node, ast.Name):
        return node.id, node.id

    # Recursive case
    # node ⊂ ( StrictlyNameable \ { ast.Name } )
    if isinstance(node, ast.Call):
        basename, sub_name = get_basename_fullname_pair(node.func, safe)
    elif isinstance(node, StrictlyNameable.__args__):
        basename, sub_name = get_basename_fullname_pair(node.value, safe)

    if isinstance(node, ast.Attribute):
        return basename, f"{sub_name}.{node.attr}"

    if isinstance(node, ast.Subscript):
        return basename, f"{sub_name}[]"

    if isinstance(node, ast.Call):
        if any(is_call_to(attr, node) for attr in PYTHON_ATTR_BUILTINS):
            return basename, ".".join(get_xattr_obj_name_pair(basename, node))

        return basename, f"{sub_name}()"

    if isinstance(node, ast.Starred):
        return basename, f"*{sub_name}"

    # Error case
    # node ⊂ Nameable ^ node ⊄ StrictlyNameable
    if safe:
        return (
            f"{LOCAL_VALUE_PREFIX}{node.__class__.__name__}",
            f"{LOCAL_VALUE_PREFIX}{node.__class__.__name__}",
        )

    _error_class: Type[TypeError] = TypeError
    if isinstance(node, ast.UnaryOp):
        _error_class = error.RattrUnaryOpInNameable
    elif isinstance(node, ast.BinOp):
        _error_class = error.RattrBinOpInNameable
    elif isinstance(node, Constant.__args__):
        _error_class = error.RattrConstantInNameable
    elif isinstance(node, Literal.__args__):
        _error_class = error.RattrLiteralInNameable
    elif isinstance(node, Comprehension.__args__):
        _error_class = error.RattrComprehensionInNameable
    else:
        _error_class = TypeError

    raise _error_class(f"line {node.lineno}: {ast.dump(node)}")


def get_basename(node: Nameable, safe: bool = False) -> str:
    """Return the `_identifier` of the innermost ast.Name node."""
    return get_basename_fullname_pair(node, safe)[0]


def get_fullname(node: Nameable, safe: bool = False) -> str:
    """Return the fullname of the given node."""
    return get_basename_fullname_pair(node, safe)[1]


def get_attrname(node: Nameable) -> str:
    """Return the `_identifier` of the outermost attribute in a nested name.

    E.g. `a.b.c.d` -> `d`
    >>> attr = ast.parse("object.attribute").body[0].value
    """
    if isinstance(node, ast.Name):
        return node.id

    if isinstance(node, ast.Attribute):
        return node.attr

    if isinstance(node, ast.Call):
        return get_attrname(node.func)

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def unravel_names(
    node: Nameable,
    *,
    get_name: Callable[[Nameable], str] = get_basename,
) -> Iterable[str]:
    """Return the basename of each nameable in the given node.

    >>> ravelled_names = ast.parse("a, b = 1, 2").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a", "b"]

    >>> ravelled_names = ast.parse("(a, b), c, d.e = ...").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a", "b", "c", "d"]

    The name getter can be overridden for example as `get_fullname`:
    >>> ravelled_names = ast.parse("a.attr = 1").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a"]
    >>> list(unravel_names(ravelled_names, get_name=get_fullname))
    ["a.attr"]

    """
    if isinstance(node, StrictlyNameable.__args__):
        return [get_name(node)]

    if isinstance(node, (ast.Tuple, ast.List)):
        ravelled = [unravel_names(i, get_name=get_name) for i in node.elts]
        return chain.from_iterable(ravelled)

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def is_call_to(target: str, node: ast.Call) -> bool:
    """Return `True` if the given node is a __direct__ call to `target`.

    In the case `getattr(o, a).some_call()`, the basename of `.some_call()` is
    `getattr` -- this is the correct basename but gives incorrect behaviour;
    the args of `getattr` are not properly resolved and `.some_call()` is not
    fully visited.

    Thus rather than determining if a call is to `getattr` by checking the
    basename against "getattr", use `is_call_to("getattr", node)`.

    """
    if not isinstance(node, ast.Call):
        raise TypeError(f"line {node.lineno}: {ast.dump(node)}")

    if not isinstance(node.func, ast.Name):
        return False

    return node.func.id == target


def has_affect(builtin: str) -> bool:
    """Return `True` if the given builtin can affect accessed attributes."""
    if builtin not in PYTHON_BUILTINS:
        raise ValueError(f"'{builtin}' is not a Python builtin")

    return builtin in PYTHON_ATTR_BUILTINS


def get_xattr_obj_name_pair(
    xattr: str, node: ast.Call, warn: bool = False
) -> Tuple[str, str]:
    """Return the object-name pair for a call to getattr, setattr, etc."""
    if len(node.args) < 2:
        error.fatal(f"invalid call to '{xattr}', not enough args", node)

    this = f"'{xattr}'"
    obj, attr = node.args[0], node.args[1]

    # NOTE Only warn if told to, avoids repeated warnings
    if warn and not isinstance(attr, ast.Str):
        error.error(f"{this} expects name to be a string literal", node)

    if isinstance(attr, ast.Str):
        attr_name = attr.s
    else:
        attr_name = f"<{get_fullname(attr, safe=True)}>"

    # Nested calls must be to the same function
    if isinstance(obj, ast.Call) and not is_call_to(xattr, obj):
        error.fatal(f"{this} object must be a name or a call to {this}", node)

    # Recurive Case
    # E.g. `getattr(getattr(a, 'b'), 'c')`
    if isinstance(obj, ast.Call):
        return ".".join(get_xattr_obj_name_pair(xattr, obj)), attr_name

    # Base Case
    # NOTE Call ∈ StrictlyNameable, thus base case comes second
    if isinstance(obj, StrictlyNameable.__args__):
        return get_fullname(node.args[0]), attr_name

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_decorator_name(decorator: ast.expr) -> str:
    """Return name of the given decorator.."""
    if not isinstance(decorator, (ast.Name, ast.Attribute, ast.Call)):
        raise TypeError("Decorator must be an ast.Name, ast.Attribute, or ast.Call")

    return get_basename(decorator)


def get_nth_argument_name(arguments: ast.arguments, n: int) -> str:
    """Return the name of the nth positional argument, or Ɛ if out-of-range."""
    _arguments = arguments.args

    if len(_arguments) == 0:
        return ""

    if not (0 <= n < len(_arguments)):
        return ""

    return _arguments[n].arg


def get_first_argument_name(arguments: ast.arguments) -> str:
    return get_nth_argument_name(arguments, 0)


def get_second_argument_name(arguments: ast.arguments) -> str:
    return get_nth_argument_name(arguments, 1)


def has_annotation(name: str, fn: AstDef) -> bool:
    """Return `True` if the function is decorated with the given annotation."""
    return name in map(get_attrname, fn.decorator_list)


def get_annotation(name: str, fn: AstDef) -> Optional[ast.expr]:
    """Return the decorator node for the given annotation on the function."""
    matching: List[ast.expr] = list()

    for decorator in fn.decorator_list:
        suffix = get_attrname(decorator)

        if suffix != name:
            continue

        matching.append(decorator)

    if len(matching) < 1:
        return None
    if len(matching) > 1:
        error.fatal(f"duplicated annotation '{name}' on '{fn.name}'", fn)

    return matching[0]


def safe_eval(expr: ast.expr, culprit: ast.AST) -> Optional[Any]:
    """Return the given expression as evaluated at compile-time.

    NOTE
        Use this instead of Python's `eval` to avoid potential arbitrary code
        execution.

    """
    _error = "unable to evaluate {} at compile-time"

    if isinstance(expr, ast.Num):
        return expr.n

    if isinstance(expr, (ast.Str, ast.Bytes)):
        return expr.s

    if isinstance(expr, ast.NameConstant):
        return expr.value

    if isinstance(expr, ast.List):
        return [safe_eval(e, culprit) for e in expr.elts]

    if isinstance(expr, ast.Tuple):
        return tuple([safe_eval(e, culprit) for e in expr.elts])

    if isinstance(expr, ast.Set):
        return set([safe_eval(e, culprit) for e in expr.elts])

    if isinstance(expr, ast.Dict):
        if not all(k is not None for k in expr.keys):
            return error.fatal(_error.format("dictionary unpacking"), culprit)

        keys = map(lambda k: safe_eval(k, culprit), expr.keys)
        values = map(lambda v: safe_eval(v, culprit), expr.values)

        return {k: v for k, v in zip(keys, values)}

    return error.fatal(_error.format(f"'{ast.dump(expr)}'"), culprit)


def parse_annotation(
    name: str, fn_def: FuncOrAsyncFunc
) -> Tuple[List[Any], Dict[str, Any]]:
    """Return the positional and keyword arguments of the annotation."""
    annotation = get_annotation(name, fn_def)

    pos_args: List[Any] = list()
    named_args: Dict[str, Any] = dict()

    if not annotation:
        return pos_args, named_args

    if not isinstance(annotation, ast.Call):
        return pos_args, named_args

    for arg in annotation.args:
        pos_args.append(safe_eval(arg, fn_def))

    for kwarg in annotation.keywords:
        named_args[kwarg.arg] = safe_eval(kwarg.value, fn_def)

    return pos_args, named_args


def is_name(name: Any) -> bool:
    """Return `True` if the given name is a valid Python `_identifier`.

    See: `parse_rattr_results_from_annotation(...)`
    """
    if not isinstance(name, str):
        return False

    if name.startswith("*"):
        name = name[1:]

    if name.startswith("@"):
        name = name[1:]

    return bool(re.match("^[A-Za-z_][A-Za-z\\d\\(\\)\\[\\]\\._]*$", name))


def is_set_of_names(set_of_names: Any) -> bool:
    """Return `True` if the given names are all valid Python `_identifier`s.

    See: `is_name(...`)
    See: `parse_rattr_results_from_annotation(...)`
    """
    if not isinstance(set_of_names, set):
        return False

    return all(is_name(name) for name in set_of_names)


def is_args(args: Any) -> bool:
    """Return `True` if the given names are all valid Python `_identifier`s.

    See: `is_name(...`)
    See: `parse_rattr_results_from_annotation(...)`
    """
    if not isinstance(args, tuple):
        return False

    if len(args) != 2:
        return False

    pos_args, named_args = args

    if not is_set_of_names(set(pos_args)):
        return False

    if not all(is_name(name) for name in named_args.keys()):
        return False
    if not all(is_name(arg) for arg in named_args.values()):
        return False

    return True


def parse_rattr_results_from_annotation(fn_def: AnyFunctionDef, context) -> FunctionIR:
    """Return the IR for the given function, assuming it is annotated."""
    # Check arguments
    expected = {"sets": set(), "gets": set(), "calls": list(), "dels": set()}
    pos_args, named_args = parse_annotation("rattr_results", fn_def)

    if pos_args != list():
        error.fatal(
            f"unexpected positional arguments to 'rattr_results'; expected "
            f"none, got {pos_args}",
            fn_def,
        )

    for name, default in expected.items():
        if name in named_args:
            continue
        named_args[name] = default

    if not all(key in expected.keys() for key in named_args.keys()):
        error.fatal(
            f"'rattr_results' expects one-or-many of the arguments "
            f"{set(expected.keys())}, found {set(named_args.keys())}",
            fn_def,
        )

    # Check argument types
    def __type_error(msg: str) -> None:
        error.fatal("type error:" + msg, fn_def)

    if not is_set_of_names(named_args.get("sets")):
        __type_error("'rattr_results' expects set of names for 'sets'")

    if not is_set_of_names(named_args.get("gets")):
        __type_error("'rattr_results' expects set of names for 'gets'")

    if not is_set_of_names(named_args.get("dels")):
        __type_error("'rattr_results' expects set of names for 'dels'")

    if not all(is_name(c[0]) for c in named_args.get("calls")):
        __type_error("'rattr_results' LHS of 'calls' to be names")

    if not all(is_args(c[1]) for c in named_args.get("calls")):
        __type_error("'rattr_results' expects args for 'calls'")

    if not all(len(c) == 2 for c in named_args.get("calls")):
        __type_error("'rattr_results' expects two elements per call")

    # Parse arguments as IR
    return {
        "sets": {parse_name(n) for n in named_args.get("sets")},
        "gets": {parse_name(n) for n in named_args.get("gets")},
        "dels": {parse_name(n) for n in named_args.get("dels")},
        "calls": {
            parse_call(n, a, context.get_call_target(n, fn_def))
            for n, a in named_args.get("calls")
        },
    }


def is_blacklisted_module(module: str) -> bool:
    """Return `True` if the given module matches a blacklisted pattern."""
    # Exclude stdlib modules such as the built-in "_thread"
    if is_stdlib_module(module):
        return False

    # Allow user specified exclusions via CLI
    blacklist = set.union(MODULE_BLACKLIST_PATTERNS, config.excluded_imports)

    return any(re.fullmatch(p, module) for p in blacklist)


def is_pip_module(module: str) -> bool:
    """Return `True` if the given module is pip installed."""
    pip_install_locations = (".+/site-packages.*",)

    try:
        spec = find_spec(module)
    except (AttributeError, ModuleNotFoundError, ValueError):
        spec = None

    if spec is None or spec.origin is None:
        return False

    # No backslashes, bad windows!
    spec.origin = spec.origin.replace("\\", "/")

    return any(re.fullmatch(p, spec.origin) for p in pip_install_locations)


def is_stdlib_module(module: str) -> bool:
    """Return `True` if the given module is in the Python standard library.

    >>> is_stdlib_module("math")
    True

    >>> is_stdlib_module("math.pi")
    True

    """
    return place_module(module) == "STDLIB"


def is_in_builtins(name_or_qualified_name: str) -> bool:
    return name_or_qualified_name in dir(builtins)


def get_function_def_args(
    fn_def: AnyFunctionDef,
) -> Tuple[List[str], Optional[str], Optional[str]]:
    """Return the arguments in the given function definition."""
    args = list()
    vararg = None
    kwarg = None

    for pos in fn_def.args.args:
        args.append(pos.arg)

    if fn_def.args.vararg:
        vararg = fn_def.args.vararg.arg

    if fn_def.args.kwarg:
        kwarg = fn_def.args.kwarg.arg

    return args, vararg, kwarg


def get_function_call_args(
    fn_call: ast.Call,
    self_name: Optional[str] = None,
) -> Tuple[List[str], Dict[str, str]]:
    """Return the arguments in the given function call.

    Method/initialiser calls can provide a `self_name` to prepend to the
    positional arguments.

    """
    _unsupported = "{} not supported in function calls"

    args = list()
    kwargs = dict()

    for arg in fn_call.args:
        if isinstance(arg, ast.Starred):
            error.error(_unsupported.format("iterable unpacking"), fn_call)

        args.append(get_fullname(arg, safe=True))

    for kwarg in fn_call.keywords:
        expected, got = kwarg.arg, kwarg.value

        if expected is None:
            error.fatal(_unsupported.format("dictionary unpacking"), fn_call)
            return list(), dict()  # Make mypy happy

        kwargs[expected] = get_fullname(got, safe=True)

    if self_name is not None:
        args = [self_name, *args]

    return args, kwargs


def remove_call_brackets(call: str):
    """Return the given string with the trailing `()` removed, if present.

    NOTE Made redundant by `str.removesuffix("()")` in Python 3.9+
    """
    if call.endswith("()"):
        return call[:-2]

    return call


class timer:
    """Context manager to return the context's elapsed time in seconds."""

    def __enter__(self):
        self.start = perf_counter()
        self.end = 0.0
        return self

    def __exit__(self, *_):
        self.end = perf_counter()

    @property
    def time(self):
        return self.end - self.start


class read:
    """Context manager to return file contents and the number of lines."""

    def __init__(self, file: str) -> None:
        self.lines = 0
        self.contents = ""
        self.file = file

    def __enter__(self):
        with open(self.file, "r") as f:
            for _lineno, line in enumerate(f):
                self.contents += line
            self.lines = _lineno + 1
        return self.lines, self.contents

    def __exit__(self, *_):
        pass


class Changes:
    """Context manager to return the dict-like's changes across the block."""

    def __init__(self, target, keys_fn=lambda t: t.keys()):
        self.target = target
        self.keys_fn = keys_fn

    def __enter__(self):
        self.antecedent: Set = set(self.keys_fn(self.target))
        return self

    def __exit__(self, *_):
        self.consequent: Set = set(self.keys_fn(self.target))

    @property
    def added(self) -> Set:
        """Return the keys added to the IR."""
        return self.consequent - self.antecedent

    @property
    def removed(self) -> Set:
        """Return the keys removed from the IR, should be the empty set."""
        return self.antecedent - self.consequent


def get_starred_imports(
    symbols: List[Symbol],
    seen: Iterable[Import],
) -> List[Import]:
    """Return the starred imports in the given symbols."""
    starred: List[Import] = list()

    for symbol in symbols:
        if not isinstance(symbol, Import):
            continue

        if symbol.name != "*":
            continue

        if symbol in seen:
            continue

        starred.append(symbol)

    return starred


@contextmanager
def enter_file(new_file: str) -> Generator:
    """Set `config.current_file` for the scope of a context."""
    old_file = config.current_file
    config.current_file = new_file
    yield
    config.current_file = old_file


def get_function_body(node: AnyFunctionDef) -> List[ast.stmt]:
    """Return the body of the given function as a list of statements."""
    if isinstance(node, ast.Lambda):
        return [node.body]

    if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
        return node.body

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_assignment_targets(node: AnyAssign) -> List[ast.expr]:
    """Return the given assignment's targets as a list."""
    if isinstance(node, ast.Assign):
        return node.targets

    if isinstance(node, (ast.AnnAssign, ast.AugAssign, ast_NamedExpr)):
        return [node.target]

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_contained_walruses(node: AnyAssign) -> List[ast_NamedExpr]:
    """Return the walruses in the RHS of the given assignment.

    >>> get_nested_walruses(ast.parse("a = (b := c)"))
    [NamedExpr(target=Name(id='b', ctx=Store()), value=Name(id='c', ctx=Load())), ]

    >>> get_nested_walruses(ast.parse("a = (b, c := d)"))
    [NamedExpr(target=Name(id='c', ctx=Store()), value=Name(id='d', ctx=Load())), ]
    """
    if not walrus_in_rhs(node):
        return list()

    if isinstance(node.value, (ast.Tuple, ast.List)):
        rhs_values = node.value.elts
    else:
        rhs_values = [node.value]

    return list(filter(lambda v: isinstance(v, ast_NamedExpr), rhs_values))


def assignment_is_one_to_one(node: AnyAssign) -> bool:
    """Return `True` if the given assignment is one-to-one."""
    _iterable = (ast.Tuple, ast.List)

    targets = get_assignment_targets(node)

    # Ensure LHS is singular
    if len(targets) > 1 or any(isinstance(t, _iterable) for t in targets):
        return False

    return not isinstance(node.value, _iterable)


def lambda_in_rhs(node: AnyAssign) -> bool:
    """Return `True` if the RHS of the given assignment contains a lambda."""
    _iterable = (ast.Tuple, ast.List)

    if isinstance(node.value, ast.Lambda):
        return True
    elif isinstance(node.value, _iterable):
        return any(isinstance(v, ast.Lambda) for v in node.value.elts)
    else:
        return False


def walrus_in_rhs(node: AnyAssign) -> bool:
    """Return `True` if the RHS contains a walrus operator."""
    _iterable = (ast.Tuple, ast.List)

    if isinstance(node.value, ast_NamedExpr):
        return True
    elif isinstance(node.value, _iterable):
        return any(isinstance(v, ast_NamedExpr) for v in node.value.elts)
    else:
        return False


def class_in_rhs(node: AnyAssign, context: Any) -> bool:
    """Return `True` if the RHS of the given assignment is a class init."""
    _iterable = (ast.Tuple, ast.List)

    def expr_is_class(expr):
        full = get_fullname(expr, safe=True)

        try:
            target = context.get_call_target(full, expr, False)
        except ValueError:
            return False

        return isinstance(target, Class)

    if isinstance(node.value, ast.Call):
        return expr_is_class(node.value)
    elif isinstance(node.value, _iterable):
        return any(expr_is_class(e) for e in node.value.elts)
    else:
        return False


def is_starred_import(node: Union[ast.Import, ast.ImportFrom]) -> bool:
    """Return `True` if the given import is a `from module import *`."""
    if not isinstance(node, ast.ImportFrom):
        return False

    return any(target.name == "*" for target in node.names)


def is_relative_import(node: Union[ast.Import, ast.ImportFrom]) -> bool:
    """Return `True` if the given import is a relative import."""
    if not isinstance(node, ast.ImportFrom):
        return False

    return node.level != 0


def module_name_from_file_path(file: str) -> Optional[str]:
    """Return the recognised name of the given module."""
    module = file.replace("/", ".").replace("\\", ".")

    if module.endswith(".__init__.py"):
        module = module[:-12]

    if module.endswith(".py"):
        module = module[:-3]

    module = module.strip(".")

    # Get possible module names in relevancy order
    parts = module.split(".")
    dotted_parts = list(chain.from_iterable(zip(parts, ["."] * len(parts))))
    possible = list(accumulate(reversed(dotted_parts), lambda p0, p1: f"{p1}{p0}"))
    well_formed = list(filterfalse(lambda p: p.startswith("."), possible))
    well_formed = list(wf[:-1] for wf in well_formed if len(wf) >= 1)
    ordered = list(reversed(well_formed))

    for p in [*ordered, None]:
        try:
            spec = find_spec(p)
            if spec is not None:
                break
        except (AttributeError, ModuleNotFoundError, ValueError):
            continue

    return p


def get_absolute_module_name(base: str, level: int, target: str) -> str:
    """Return the absolute import for the given relative import."""
    level -= int(config.current_file.endswith("__init__.py"))

    if level > 0:
        new_base = ".".join(base.split(".")[:-level])
    else:
        new_base = base

    return f"{new_base}.{target}"


def get_function_form(node: AnyFunctionDef) -> str:
    """Return the string representing the function form.

    Given the function, fn:
        ```
        def fn(a, b, c=None, *args, **kwargs):
            # ...
        ```

    The form is:
        `fn(a, b, c, *args, **kwargs)`

    """
    arguments = [a.arg for a in node.args.args]

    if node.args.vararg is not None:
        arguments.append(f"*{node.args.vararg.arg}")

    if node.args.kwarg is not None:
        arguments.append(f"**{node.args.kwarg.arg}")

    return f"'{node.name}({', '.join(arguments)})'"


def re_filter_ir(d: Dict[Symbol, Any], filter_by: str) -> Dict[str, Any]:
    """Return the given dict filtered by the given regular expression."""
    if not filter_by:
        return d

    return {f: d[f] for f in filter(lambda f: re.fullmatch(filter_by, f.name), d)}


def re_filter_results(d: Dict[str, Any], filter_by: str) -> Dict[str, Any]:
    """Return the given dict filtered by the given regular expression."""
    if not filter_by:
        return d

    return {f: d[f] for f in filter(lambda f: re.fullmatch(filter_by, f), d)}


def is_excluded_name(name: str) -> bool:
    """Return `True` if the given name is in the exclude list."""
    return any(re.fullmatch(x, name) for x in config.excluded_names)


def is_method_on_constant(name: str) -> bool:
    """Return `True` if the name is a call to method on a constant."""
    if sys.version_info.major == 3 and sys.version_info.minor <= 7:
        return name.startswith(
            (
                "@Num.",
                "@Str.",
                "@Bytes.",
                "@NameConstant.",
            )
        )

    if sys.version_info.major == 3 and sys.version_info.minor >= 8:
        return name.startswith("@Constant.")

    raise NotImplementedError


def is_method_on_cast(name: str) -> bool:
    """Return `True` if the name is a call to a method on a cast."""
    # NOTE All builtin functions return primitives (often `None`), ...
    builtins = {b for b in PYTHON_BUILTINS if b.startswith((*ascii_lowercase,))}

    # NOTE ... with the following exceptions
    builtins = builtins.difference(
        {
            "eval",
            "exec",
            "getattr",
            "iter",
            "next",
            "slice",
            "super",
            "type",
        }
    )

    # NOTE
    #   Direct call to `<primitive_returner>.method_name()`, do not allow call
    #   to call, etc because one may return non-primitive
    patterns = tuple((f"{b}\\(\\)\\.[^\\(\\)\\[\\]\\.]+" for b in builtins))

    return any(re.fullmatch(p, name) for p in patterns)


def is_method_on_primitive(name: str) -> bool:
    """Return `True` if the given name is a method on a primitive type."""
    if is_method_on_constant(name):
        return True

    if is_method_on_cast(name):
        return True

    return False


def get_dynamic_name(fn_name: str, node: ast.Call, pattern: str) -> Name:
    """Return the `Name` of the given pattern constructed from the call.

    Where first is the resolved name of the first call argument, and second is
    the resolved name of the second call argument.

    Assumes that the first argument is a variable (or nested call) and the
    second argument is a string -- names are resolved as such.

    Allows nested calls to functions with the same name (i.e. `name`).

    >>> call = ast.parse("getattr(object, 'attribute')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    Name("object.attribute", "object")

    >>> call = ast.parse("getattr(getattr(a, 'b'), 'c')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    Name("a.b.c", "a")

    >>> call = ast.parse("getattr(another_fn(a, 'b'), 'c')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    sys.exit(1)

    >>> call = ast.parse("get_sub_attr(obj, 'attr')").body[0].value
    >>> get_dynamic_name("get_sub_attr", call, "{first}.sub.{second}")
    Name("obj.sub.attr", "obj")

    >>> call = ast.parse("get_sub_attr(get_sub_attr(o, 'in'), 'out')").body[0].value
    >>> get_dynamic_name("get_sub_attr", call, "{first}.mid.{second}")
    Name("o.in.mid.out", "o")

    >>> call = ast.parse("get_sub(obj)").body[0].value
    >>> get_dynamic_name("get_sub", call, "{first}.sub")
    Name("obj.sub", "obj")

    """
    first, second = get_xattr_obj_name_pair(fn_name, node, warn=True)

    basename = first.split(".")[0].replace("*", "").replace("[]", "").replace("()", "")

    return Name(pattern.format(first=first, second=second), basename)


def get_file_hash(filepath: str, blocksize: int = 2**20) -> str:
    """Return the hash of the given file, with a default blocksize of 1MiB."""
    _hash = hashlib.md5()

    if not isfile(filepath):
        return _hash

    with open(filepath, "rb") as f:
        while True:
            buffer = f.read(blocksize)

            if not buffer:
                break

            _hash.update(buffer)

    return _hash.hexdigest()


def cache_is_valid(filepath: str, cache_filepath: str) -> bool:
    """Return `True` if the cache has the correct hash."""
    if not isfile(filepath):
        return False

    if isfile(cache_filepath):
        with open(cache_filepath, "r") as f:
            cache: Dict[str, Any] = json.load(f)
    else:
        return False

    received_hash = cache.get("filehash", None)
    expected_hash = get_file_hash(filepath)

    if filepath != cache.get("filepath", None):
        return False

    if received_hash != expected_hash:
        return False

    # Check imports
    for _import in cache.get("imports", dict()):
        file, hash = _import["filepath"], _import["filehash"]
        if hash != get_file_hash(file):
            return False

    return True


def create_cache(
    results: FileResults, imports: Set[str], encoder: json.JSONEncoder
) -> None:
    """Create a the cache and write it to the cache file.

    NOTE:
        The attribute `imports` should hold the file name of every directly and
        indirectly imported source code file.

    Cache file format (JSON):
    {
        "filepath": ...,    # the file the results belong to
        "filehash": ...,    # the MD5 hash of the file when it was cached

        "imports": [
            {"filename": ..., "filehash": ...,},
        ]

        "results":
            # For each function in the file:
            "function_name": {
                "sets"  : ["obj.attr", ...],
                "gets"  : ["obj.attr", ...],
                "dels"  : ["obj.attr", ...],
                "calls" : ["obj.attr", ...],
            },
            ...
        }
    }

    """
    to_cache = dict()

    to_cache["filepath"] = config.file
    to_cache["filehash"] = get_file_hash(config.file)
    to_cache["imports"] = [
        {"filepath": file, "filehash": get_file_hash(file)} for file in imports
    ]
    to_cache["results"] = deepcopy(results)

    with open(config.cache, "w") as f:
        json.dump(to_cache, f, cls=encoder, indent=4)



"""Rattr file analyser."""

import ast
from collections import namedtuple
from copy import deepcopy
from dataclasses import replace as copy_dataclass
from typing import List, NamedTuple, Set, Tuple

from rattr import config, error
from rattr.analyser.base import NodeVisitor
from rattr.analyser.cls import ClassAnalyser
from rattr.analyser.context import Context, Func, Import, RootContext
from rattr.analyser.function import FunctionAnalyser
from rattr.analyser.types import (
    AnyAssign,
    AnyFunctionDef,
    FileIR,
    ImportsIR,
    ast_NamedExpr,
)
from rattr.analyser.util import (
    Changes,
    assignment_is_one_to_one,
    enter_file,
    get_assignment_targets,
    get_contained_walruses,
    get_fullname,
    has_annotation,
    is_blacklisted_module,
    is_excluded_name,
    is_pip_module,
    is_stdlib_module,
    lambda_in_rhs,
    parse_rattr_results_from_annotation,
    read,
    timer,
)
from rattr.plugins import plugins

RattrStats = namedtuple(
    "RattrStats",
    [
        "parse_time",
        "root_context_time",
        "assert_time",
        "analyse_imports_time",
        "analyse_file_time",
        "file_lines",
        "import_lines",
        "number_of_imports",
        "number_of_unique_imports",
    ],
)

ImportStats = namedtuple(
    "ImportStats",
    [
        "import_lines",
        "number_of_imports",
        "number_of_unique_imports",
    ],
)


def parse_and_analyse_file() -> Tuple[FileIR, ImportsIR, NamedTuple]:
    """Parse and analyse `config.file`."""
    with enter_file(config.file):
        file_ir, imports_ir, stats = __parse_and_analyse_file()

    return file_ir, imports_ir, stats


def __parse_and_analyse_file() -> Tuple[FileIR, ImportsIR, NamedTuple]:
    """Parse and analyse the given file contents."""
    with timer() as parse_timer, read(config.file) as (file_lines, source):
        _ast = ast.parse(source)

    with timer() as root_context_timer:
        context = RootContext(_ast).expand_starred_imports()

    with timer() as assert_timer:
        for assertor in plugins.assertors:
            assertor.assert_holds(_ast, deepcopy(context))

    with timer() as analyse_imports_timer:
        if config.follow_imports:
            symbols = context.symbol_table.symbols()
            imports = list(filter(lambda s: s._is(Import), symbols))
            imports_ir, import_stats = parse_and_analyse_imports(imports)
        else:
            imports_ir, import_stats = dict(), ImportStats(0, 0, 0)

    with timer() as analyse_file_timer:
        file_ir = FileAnalyser(_ast, context).analyse()

    stats = RattrStats(
        parse_timer.time,
        root_context_timer.time,
        assert_timer.time,
        analyse_imports_timer.time,
        analyse_file_timer.time,
        file_lines,
        import_stats.import_lines,
        import_stats.number_of_imports,
        import_stats.number_of_unique_imports,
    )

    return file_ir, imports_ir, stats


def parse_and_analyse_imports(imports: List[Import]) -> Tuple[ImportsIR, ImportStats]:
    """Return the mapping from file name to IR for each import.

    Imports are a directed cyclic graph, however, previously analysed files can
    just be ignored (analysing is deterministic and context-free). Thus, the
    graph of imports becomes a DAG which we BFS.

    """
    n_lines: int = 0
    n_imports: int = 0
    imports_ir: ImportsIR = dict()
    seen_module_paths: Set[str] = set()

    for _i in imports:
        module_name = _i.module_name
        module_path = _i.module_spec.origin

        if module_name is None or module_path is None:
            # HACK Can't use isinstance, TODO Resolve BuiltinImporter modules
            if "BuiltinImporter" in str(_i.module_spec.loader):
                error.error(
                    f"unable to resolve builtin module '{module_name}'", badness=0
                )
                continue

            error.fatal(f"unable to resolve import '{_i.qualified_name}'")
            continue

        if module_path in seen_module_paths:
            continue

        if is_blacklisted_module(module_name):
            continue

        if not config.follow_pip_imports and is_pip_module(module_name):
            continue

        if not config.follow_stdlib_imports and is_stdlib_module(module_name):
            continue

        with read(module_path) as (file_lines, source):
            _i_ast = ast.parse(source)

        with enter_file(module_path):
            _i_ctx = RootContext(_i_ast).expand_starred_imports()
            _i_ir = FileAnalyser(_i_ast, _i_ctx).analyse()

        imports_ir[module_name] = _i_ir

        # Add the import's imports to the queue
        _i_imports = list(
            filter(lambda s: s._is(Import), _i_ctx.symbol_table.symbols())
        )
        imports += _i_imports

        n_lines += file_lines
        n_imports += len(_i_imports)

        seen_module_paths.add(module_path)

    return imports_ir, ImportStats(n_lines, n_imports, len(seen_module_paths))


class FileAnalyser(NodeVisitor):
    """Walk a file's AST and analyse the contained functions and classes."""

    def __init__(self, _ast: ast.Module, context: Context) -> None:
        """Set configuration and initialise results."""
        self._ast = _ast
        self.context = context
        self.file_ir = FileIR(context)

    def analyse(self) -> FileIR:
        """Entry point of FileAnalyser, return the results of analysis."""
        self.visit(self._ast)

        return self.file_ir

    def visit_AnyFunctionDef(self, node: AnyFunctionDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if is_excluded_name(node.name):
            return

        fn: Func = self.context.get(node.name)

        if has_annotation("rattr_results", node):
            self.file_ir[fn] = parse_rattr_results_from_annotation(node, self.context)
            return

        if plugins.custom_function_handler.has_analyser(node.name, self.context):
            self.file_ir[fn] = plugins.custom_function_handler.get(
                node.name, self.context
            ).on_def(node.name, node, self.context)
            return

        self.file_ir[fn] = FunctionAnalyser(node, self.context).analyse()

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if is_excluded_name(node.name):
            return

        class_ir = ClassAnalyser(node, self.context).analyse()

        for foc, foc_ir in class_ir.items():
            self.file_ir[foc] = foc_ir

    # ----------------------------------------------------------------------- #
    # Lambdas
    # ----------------------------------------------------------------------- #

    def visit_LambdaAssign(self, node: AnyAssign) -> None:
        if not assignment_is_one_to_one(node):
            return error.fatal("lambda assignment must be one-to-one", node)

        targets = get_assignment_targets(node)
        name = get_fullname(targets[0])

        fn, context = self.context.get(name), self.context
        self.file_ir[fn] = FunctionAnalyser(node.value, context).analyse()

    def visit_AnyAssign(self, node: AnyAssign) -> None:
        if lambda_in_rhs(node):
            self.visit_LambdaAssign(node)

        # Walrus may obscure a lambda, so peek in and visit the nice walruses
        for walrus in get_contained_walruses(node):
            with Changes(self.file_ir) as diff:
                self.visit_AnyAssign(walrus)

            # If the walrus is of the form "a = (b := lambda ...)" then "a" should have
            # the same result as "b" which would have just been registered
            if lambda_in_rhs(walrus):
                if len(diff.added) == 1 and node.value == walrus:
                    rhs: Func = list(diff.added)[0]
                    lhs: Func = copy_dataclass(
                        rhs, name=get_fullname(get_assignment_targets(node)[0])
                    )
                    self.file_ir[lhs] = self.file_ir[rhs]
                elif len(diff.added) > 1:
                    raise NotImplementedError("Multiple deeply nested walruses")

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast_NamedExpr) -> None:
        self.visit_AnyAssign(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        # NOTE Only reached when the lambda is an anonymous function
        return error.fatal("module level lambdas unsupported", node)



"""Rattr types."""

import ast
import json
import sys
from typing import (
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    Optional,
    Set,
    Type,
    Union,
    ValuesView,
)

from rattr.analyser.context.symbol import Class, Func, Symbol

Constant = Union[
    ast.Constant,
    # Deprecated in Python 3.8, replaced with ast.Constant(..., kind=?)
    ast.Num,
    ast.Str,
    ast.Bytes,
    ast.NameConstant,
    ast.Ellipsis,
]

Literal = Union[
    ast.FormattedValue,
    ast.JoinedStr,
    ast.List,
    ast.Tuple,
    ast.Set,
    ast.Dict,
]

Comprehension = Union[
    ast.ListComp,
    ast.SetComp,
    ast.GeneratorExp,
    ast.DictComp,
]


Nameable = ast.expr

# Nodes which contain a name attribute (.func, .id, .attr, etc)
StrictlyNameable = Union[
    ast.Name,
    ast.Attribute,
    ast.Subscript,
    ast.Starred,
    ast.Call,
]

# The subset of `StrictlyNameable` which contain `Name` as their base element
CompoundStrictlyNameable = Union[
    ast.Attribute,
    ast.Subscript,
    ast.Starred,
]

AstDef = Union[
    ast.FunctionDef,
    ast.AsyncFunctionDef,
    ast.ClassDef,
]

# Union of Python literals
PythonLiteral = Union[
    None,
    int,
    float,
    complex,
    bool,
    str,
    bytes,
]

FuncOrAsyncFunc = Union[ast.FunctionDef, ast.AsyncFunctionDef]

AnyFunctionDef = Union[
    ast.FunctionDef,
    ast.AsyncFunctionDef,
    ast.Lambda,
]


AnyAssign = Union[
    ast.Assign,
    ast.AnnAssign,
    ast.AugAssign,
]


# --------------------------------------------------------------------------- #
# Multi-version support
# --------------------------------------------------------------------------- #


class _TypeNotInPythonVersion:
    pass


# Walrus operator typing
if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    ast_NamedExpr: Type[ast.NamedExpr] = ast.NamedExpr
    AnyAssign = Union[ast.Assign, ast.AnnAssign, ast.AugAssign, ast.NamedExpr]
else:
    ast_NamedExpr = _TypeNotInPythonVersion

# --------------------------------------------------------------------------- #
# Results and intermediate representation (IR)
# --------------------------------------------------------------------------- #

FuncOrClass = Union[Func, Class]

# LHS: "sets" | "gets" | "calls" | "dels"
FunctionIR = Dict[str, Set[Symbol]]


ClassIR = Dict[FuncOrClass, FunctionIR]


_FileIR = Dict[FuncOrClass, FunctionIR]


class FileIR:
    """Provide a `Mapping[Func, FunctionIR]`, with context."""

    def __init__(self, context):
        # `self.context: Context` omitted to avoid circular import, hopefully
        # mypy's inference is sufficient
        self.context = context
        self._file_ir: _FileIR = dict()

    def __eq__(self, other) -> bool:
        if isinstance(other, dict):
            return other == self._file_ir

        if isinstance(other, FileIR):
            return other.context == self.context and other._file_ir == self._file_ir

        raise TypeError(f"Cannot compare 'FileIR' and '{type(other)}'")

    def __iter__(self) -> Iterator[FuncOrClass]:
        return iter(self._file_ir)

    def __contains__(self, item: FuncOrClass) -> bool:
        return item in self._file_ir

    def __setitem__(self, item: FuncOrClass, value: FunctionIR) -> None:
        self._file_ir[item] = value

    def __getitem__(self, item: FuncOrClass) -> FunctionIR:
        return self._file_ir[item]

    def get(
        self, item: FuncOrClass, default: Optional[FunctionIR] = None
    ) -> Optional[FunctionIR]:
        return self._file_ir.get(item, default)

    def keys(self) -> KeysView[FuncOrClass]:
        return self._file_ir.keys()

    def values(self) -> ValuesView[FunctionIR]:
        return self._file_ir.values()

    def items(self) -> ItemsView[FuncOrClass, FunctionIR]:
        return self._file_ir.items()

    class JsonEncoder(json.JSONEncoder):
        """JSON encoder for FileIR."""

        def default(self, obj):
            if isinstance(obj, set):
                return list(obj)
            if isinstance(obj, Symbol):
                return repr(obj)
            return super().default(obj)

    def __str__(self) -> str:
        cleaned = {repr(k): v for k, v in self._file_ir.items()}

        return json.dumps(cleaned, indent=4, cls=FileIR.JsonEncoder)

    def __repr__(self) -> str:
        return f"FileIR(context={self.context}, _file_ir={self._file_ir}"


# LHS: module fully-qualified name
ImportsIR = Dict[str, FileIR]


# LHS: "sets" | "gets" | "calls" | "dels"
FunctionResults = Dict[str, Set[str]]


# LHS: function name
FileResults = Dict[str, FunctionResults]



from dataclasses import dataclass, field
from typing import Optional, Set


@dataclass
class Config:
    """Hold Rattr run-time configuration.

    NOTE
        Defaults only used in tests, `__main__.py` will always override

    """

    # Import following
    follow_imports: int = 1
    follow_pip_imports: bool = False
    follow_stdlib_imports: bool = False

    # Exclusions
    excluded_imports: Set[str] = field(default_factory=set)
    excluded_names: Set[str] = field(default_factory=set)

    # Warnings
    show_warnings: bool = True
    show_imports_warnings: bool = True
    show_low_priority_warnings: bool = False

    # File name in error messages
    show_path: bool = True
    use_short_path: bool = True

    # Results and output
    show_ir: bool = False
    show_results: bool = True
    show_stats: bool = False
    silent: bool = False

    # Error threshold
    strict: bool = False
    permissive: bool = True
    threshold: int = 0
    file_badness: int = 0
    import_badness: int = 0
    simplify_badness: int = 0

    # Results filtering
    filter_string: str = ""

    # File info
    file: str = "<no_file>"

    # HACK State, cleaner than passing [str | None] to dozens of functions
    # NOTE See `util.py::enter_file`
    current_file: Optional[str] = None

    # Cache file
    cache: str = ""


config = Config()
